# Imports.

import os
import glob
import re
import numpy as np
import pandas as pd

from .. import Global_Utilities as gu

# Function definitions.

def read_raw_data_file_1( filename, f, resin_data, file_data, data ):

    pattern = re.compile( r"^Resin(\d+)_(\d+)_" )

    resin = int( pattern.search( f ).groups()[0] )

    specimen = int( pattern.search( f ).groups()[1] )

    sheet = "Ramp 10.00 °Cmin to 600.00 °C"

    df = pd.read_excel( filename, sheet )

    for i in range( len( df.columns ) ):

        df.rename( columns = {df.columns[i]:df.loc[0][i]}, inplace = True )

    df.drop( [0, 1], inplace = True )

    for i in range( len( df.columns ) ):

        data[i].append( df.iloc[:, i].to_numpy( dtype = np.float64 ) )

    if len( df.columns ) < 5:

        for i in range( 5 - len( df.columns ) ):

            data[4 - i].append( np.array( [] ) )

    file_data.append( [resin, specimen, resin_data.loc[resin]["Label"] + ".{}".format( specimen ), ""] )

def read_raw_data_file_2( filename, f, resin_data, file_data, data ):

    pass

def extract_raw_data( directory, data_directory ):
    '''Extract the raw data from the files.'''

    resin_data = gu.get_list_of_resins_data( directory ) # Obtain the spreadsheet of data for the resins.

    resins = sorted( [os.path.basename( path ) for path in glob.glob( data_directory + "*" )], key = gu.sort_raw_files_1 )

    file_data, data = [], [[], [], [], [], []]

    pattern = re.compile( r"^Resin(\d+)" )

    for r in resins:

        filenames = sorted( [os.path.basename( path ) for path in glob.glob( data_directory + r + "/*" )], key = gu.sort_raw_files_2 )

        resin = int( pattern.search( r ).groups()[0] )

        for f in filenames:

            if resin == 40 or resin == 41:

                read_raw_data_file_2( data_directory + r + "/" + f, f, resin_data, file_data, data )

            else:

                read_raw_data_file_1( data_directory + r + "/" + f, f, resin_data, file_data, data )

    return file_data, data

def standardise_data( data ):
    '''Standardise data.'''

    standard_temp_list = np.linspace( 30, 590, 5600 ).tolist()

    for i in range( len( data[0] ) ):

        mask = []

        array = data[1][i]

        for j in range( len( standard_temp_list ) ):

            interval_mask = np.where( array >= standard_temp_list[j] )[0]

            if interval_mask.any():

                mask.append( interval_mask[0] )

            else:

                mask.append( len( array ) - 1 )

        for j in range( len( data ) ):

            if j == 1:

                continue

            if data[j][i].any():

                data[j][i] = data[j][i][mask]

        print( i )

    data[1] = np.array( standard_temp_list )

def add_description_to_file_data( file_data ):
    '''Add descriptions in the form of letters to each specimen.'''

    # Add f for specimens whose main decomposition occurs much later, presumable an experimental error.
    # Add s for specimens whose experimental length is too short.

    specimens = {7:[1], 8:[1], 9:[1], 10:[1], 11:[1], 12:[1], 13:[1], 14:[1], 15:[1], 16:[1]}

    for f in file_data:

        if f[0] in specimens.keys():

            if f[1] in specimens[f[0]]:

                f[3] = f[3] + "f"

    specimens = {5:[1], 6:[1]}

    for f in file_data:

        if f[0] in specimens.keys():

            if f[1] in specimens[f[0]]:

                f[3] = f[3] + "s"

def read_files_and_preprocess( directory, data_directory, merge_groups ):
    '''Read files and preprocess data.'''

    file_data, data = extract_raw_data( directory, data_directory )

    standardise_data( data )

    add_description_to_file_data( file_data )

    if merge_groups:

        gu.merge( file_data )

    return file_data, data

def write_csv( output_directory, file_data, data, name_appendage = "" ):
    '''Write read and preprocessed data to a .csv file.'''

    for i, f in enumerate( file_data ):

        array = np.array( data[0][i] )[:, np.newaxis]

        for j in range( 1, 5 ):

            if j == 1:

                array = np.hstack( (array, np.array( data[j] )[:, np.newaxis]) )

            elif data[j][i].any():

                array = np.hstack( (array, np.array( data[j][i] )[:, np.newaxis]) )

        np.savetxt( output_directory + "TGA/Condensed_Data/Resin{}_{}_" + name_appendage + ".csv".format( f[0], f[1] ), array, delimiter = ",", fmt = "%.7f" )

    array = np.array( file_data )

    np.savetxt( output_directory + "TGA/File_Data/file_data" + name_appendage + ".csv", array, delimiter = ",", fmt = "%s" )

def read_csv( directory, output_directory, merge_groups, name_appendage = "" ):
    '''Read the preprocessed .csv files.'''

    resin_data = gu.get_list_of_resins_data( directory ) # Obtain the spreadsheet of data for the resins.

    file_data = []

    df = pd.read_csv( output_directory + "TGA/File_Data/file_data" + name_appendage + ".csv", sep = ",", header = None )

    for i in range( len( df.index ) ):

        resin = df.iloc[i, 0]
        specimen = df.iloc[i, 1]

        file_data.append( [resin, specimen, resin_data.loc[resin]["Label"] + ".{}".format( specimen ), ""] )

    data = [[], [], [], [], []]

    filenames = sorted( [os.path.basename( path ) for path in glob.glob( output_directory + "TGA/Condensed_Data/*" ) if ("_" + name_appendage + ".csv") in path], key = gu.sort_raw_files_3 )

    for f in filenames:

        df = pd.read_csv( output_directory + "TGA/Condensed_Data/" + f, sep = ",", header = None )

        for i in range( len( df.columns ) ):

            data[i].append( df.iloc[:, i].to_numpy( dtype = np.float64 ) )

        if len( df.columns ) < 5:

            for i in range( 5 - len( df.columns ) ):

                data[4 - i].append( np.array( [] ) )

    data[1] = data[1][0]

    add_description_to_file_data( file_data )

    if merge_groups:

        gu.merge( file_data )

    return file_data, data

def remove_files( file_data, data, descriptors_to_remove = "" ):
    '''Remove files not needed/wanted for analysis by searching for letters in file descriptions.'''

    files_to_remove = []

    for i in range( len( file_data ) ):

        s = file_data[i][3]

        for l in descriptors_to_remove:

            if s.find( l ) > -0.5:

                files_to_remove.append( i )

                break

    files_to_remove.reverse()

    for r in files_to_remove:

        file_data.pop( r )
        data[0].pop( r )
        data[2].pop( r )
        data[3].pop( r )
        data[4].pop( r )

    return file_data, data

def compute_mean( output_directory, file_data, data, name_appendage = "" ):
    '''Compute the mean data for each resin.'''

    m = gu.sample_mean( file_data, data[3] )

    array = m[0][:, np.newaxis]

    for i in range( 1, len( m ) ):

        array = np.hstack( (array, m[i][:, np.newaxis]) )

    np.savetxt( output_directory + "TGA/Mean_Data/Means" + name_appendage + ".csv", array, delimiter = ",", fmt = "%.7f" )

def read_mean( output_directory, data, name_appendage = "" ):
    '''Read the computed means for each resin from a file.'''

    m = []

    df = pd.read_csv( output_directory + "TGA/Mean_Data/Means" + name_appendage + ".csv", sep = ",", header = None )

    for i in range( len( df.columns ) ):

        m.append( df.iloc[:, i].to_numpy( dtype = np.float64 ) )

    if len( data ) > 5: # A previous list of means is already appended.

            data[5] = m

    else:

        data.append( m )

    return data



# Imports.

import numpy as np
import matplotlib.pyplot as plt
import matplotlib as mpl

from .. import Global_Utilities as gu

# Function definitions.

def plot_crossover_points( directory, output_directory, resin_data, file_data, sample_array, samples_to_plot, specimens, mean, data_extraction_bool, savefig ):

    feature_names, features = gu.csv_to_df_to_array_and_column_titles( output_directory + "Rheology/Features/Features.csv" )

    features_df = gu.array_with_column_titles_to_df( features, feature_names )

    specimen_mask = gu.produce_mask( sample_array, samples_to_plot )

    features_df_copy = features_df.iloc[specimen_mask, :]
    file_data_mask = np.array( file_data )[specimen_mask] # Orders file data according to specimen mask.
    file_data_mask = [[int( f[0] ), int( f[1] ), f[2], f[3]] for f in file_data_mask] # Converts file data so that f[0], etc. are integers and not np.str.
    sample_array_copy = sample_array[specimen_mask]

    if specimens:

        gu.plot_scatterplot_of_two_features( directory, features_df_copy["Rhe_Crossover"].apply( lambda x: 10 ** x ), features_df_copy["Rhe_SMCrossover"].apply( lambda x: 10 ** x ), sample_array_copy, [f[2] for f in file_data_mask], line_of_best_fit = False, xlog = True, ylog = True, xlabel = "Angular Frequency [rad/s]", ylabel = "Storage/Loss Modulus [Pa]", savefig = savefig, filename = output_directory + "Rheology/Plots/Crossover_Point_Specimen.pdf" )

    if mean:

        mean_features = gu.extract_mean_features( features_df.to_numpy(), sample_array, samples_to_plot )

        mean_features_df = gu.array_with_column_titles_to_df( mean_features, features_df_copy.columns )

        gu.plot_scatterplot_of_two_features( directory, mean_features_df["Rhe_Crossover"].apply( lambda x: 10 ** x ), mean_features_df["Rhe_SMCrossover"].apply( lambda x: 10 ** x ), sample_array_copy, [resin_data.loc[i]["Label"] for i in samples_to_plot], line_of_best_fit = False, xlog = True, ylog = True, xlabel = "Angular Frequency [rad/s]", ylabel = "Storage/Loss Modulus [Pa]", savefig = savefig, filename = output_directory + "Rheology/Plots/Crossover_Point_Mean.pdf" )

def plot_van_Gurp_Palmen_plot( ip, resin_data, file_data, data, splits, sample_array, samples_present_array, samples_to_plot, specimens, mean, colours, data_extraction_bool, savefig ):

    shiny_de = []

    data_extraction = []

    lower_bound, upper_bound = splits[0], splits[1]

    if specimens:

        for i in samples_to_plot:

            mask = np.where( sample_array == i )[0]

            for ind, j in enumerate( mask ):

                xaxis_mask = np.where( (data[3][j] * data[0] <= upper_bound) & (data[3][j] * data[0] >= lower_bound) )[0]

                # deriv = gu.derivative( np.log( data[3][j] * data[0] ), np.arctan( data[4][j] ) )

                plt.scatter( data[3][j] * data[0], data[4][j], label = file_data[j][2], color = colours[i] )

                shiny_de.append( (data[3][j][xaxis_mask] * data[0][xaxis_mask]).tolist() )
                shiny_de.append( data[4][j][xaxis_mask].tolist() )

                # data_extraction.append( data[3][j] * data[0] )
                # data_extraction.append( data[4][j] )

                # plt.scatter( data[3][j][1:-1] * data[0][1:-1], deriv, label = file_data[j][2], color = colours[i] )

        ax = plt.gca()
        ax.set_xscale( 'log' )
        plt.xlabel( "Complex Modulus [Pa]" )
        plt.ylabel( "Phase Angle [°]" )
        # plt.legend( ncol = 3, loc = 'lower right', borderaxespad = 0 )
        plt.tight_layout()

        if savefig:

            plt.savefig( ip.output_directory + "Rheology/Plots/vGP_Specimens.pdf" )

        else:

            plt.show()

        plt.close()

    if mean:

        for i in samples_to_plot:

            index = np.where( samples_present_array == i )[0][0]

            xaxis_mask = np.where( (data[7][index] * data[0] <= upper_bound) & (data[7][index] * data[0] >= lower_bound) )[0]

            plt.scatter( data[7][index][xaxis_mask] * data[0][xaxis_mask], np.arctan( data[8][index][xaxis_mask] ) * 180 / np.pi, label = resin_data.loc[i]["Label"], color = colours[i] )

            shiny_de.append( (data[7][index][xaxis_mask] * data[0][xaxis_mask]).tolist() )
            shiny_de.append( (np.arctan( data[8][index][xaxis_mask] ) * 180 / np.pi).tolist() )

            data_extraction.append( data[7][index][xaxis_mask] * data[0][xaxis_mask] )
            data_extraction.append( np.arctan( data[8][index][xaxis_mask] ) * 180 / np.pi )

        ax = plt.gca()
        ax.set_xscale( 'log' )
        plt.xlabel( "Complex Modulus [Pa]" )
        plt.ylabel( "Phase Angle [°]" )
        plt.legend( ncol = 3, bbox_to_anchor = ( 0.880, 0 ), loc = 'lower center', borderaxespad = 0 )
        plt.tight_layout()

        if savefig:

            plt.savefig( ip.output_directory + "Rheology/Plots/vGP_Means.pdf" )

        else:

            plt.show()

        plt.close()

        if ip.shiny:

            return shiny_de

        if data_extraction_bool:

            array = data_extraction[0][:, np.newaxis]

            for i in range( 1, len( data_extraction ) ):

                array = np.hstack( (array, data_extraction[i][:, np.newaxis]) )

            np.savetxt( ip.output_directory + "Plot_Coords/Unnamed.txt", array )

def plot_data( ip, file_data, data, first_derivative_data, second_derivative_data, savefig = False ):

    perform_plot_crossover_points = False

    perform_van_Gurp_Palmen_plot = True

    perform_custom_plot = False

    resin_data = gu.get_list_of_resins_data( ip.directory )

    sample, sample_array, samples_present, samples_present_array = gu.sample_data_from_file_data( file_data )

    samples_to_plot = [11, 14, 10, 4, 13, 21, 23, 18, 22, 20, 2, 3, 17, 16, 19, 1, 15, 12, 6, 5, 7, 9, 8, 24]
    samples_to_plot = [16, 17, 19, 24, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 18, 20, 21, 22, 23]
    samples_to_plot = [16, 17, 4, 8, 9, 23]

    specimens = False
    all_specimens = True
    specimen_mask_by_index = [0]
    specimen_mask = []

    mean = True

    deriv0 = True
    deriv1 = False
    deriv2 = False

    split = False
    num_splits = 2
    split_length = 120
    splits = [split_length * (i + 30 / 120) for i in range( num_splits )]
    splits = [0.1, 0.5]

    log_graph = True

    radial_graph = False

    x, y = 0, 3 # 0:af, 1:sm, 2:lm, 3:cv, 4:lf, 5-8 same but means.

    if not split:

        if perform_van_Gurp_Palmen_plot:

            splits = [0, 500000]

        else:

            splits = [int( data[0][len( data[0] ) - 1] ), int( data[0][0] )]

    if ip.shiny:

        samples_to_plot = ip.shiny_samples_to_plot

        if type( samples_to_plot ) == int:

            samples_to_plot = [samples_to_plot]

        specimen_mask = ip.shiny_specimens_to_plot

        if type( specimen_mask ) == int:

            specimen_mask = [specimen_mask]

        splits = ip.shiny_split
        mean = ip.shiny_mean
        specimens = ip.shiny_specimen
        all_specimens = False
        perform_plot_crossover_points = False

        if ip.shiny_vgp == True:

            perform_van_Gurp_Palmen_plot = True
            perform_custom_plot = False

        else:

            perform_van_Gurp_Palmen_plot = False
            perform_custom_plot = True

            if ip.shiny_cv_vs_af == True:

                if specimens:

                    x, y = 0, 3

                else:

                    x, y = 0, 7

            elif ip.shiny_pa_vs_cv == True:

                if specimens:

                    x, y = 3, 4

                else:

                    x, y = 7, 8

        radial_graph = False
        log_graph = False

    colours = gu.read_list_of_colours( ip.directory )

    shiny_de = []

    data_extraction_bool = True

    if perform_plot_crossover_points:

        plot_crossover_points( ip.directory, ip.output_directory, resin_data, file_data, sample_array, samples_to_plot, specimens, mean, data_extraction_bool, savefig )

    if perform_van_Gurp_Palmen_plot:

        return plot_van_Gurp_Palmen_plot( ip, resin_data, file_data, data, splits, sample_array, samples_present_array, samples_to_plot, specimens, mean, colours, data_extraction_bool, savefig )

    if perform_custom_plot:

        for s in range( len( splits ) - 1 ):

            data_extraction = []

            lower_bound, upper_bound = splits[s], splits[s + 1]

            if radial_graph:

                ax = plt.subplots( 1, 1, subplot_kw = dict( polar = True ) )[1]

                for i in samples_to_plot:

                    index = np.where( samples_present_array == i )[0][0]

                    freq_mask = np.where( (data[0] <= upper_bound) & (data[0] >= lower_bound) )[0]

                    alpha_scale = np.linspace( 0.3, 1, len( data[8][index][freq_mask] ) )

                    ax.scatter( np.arctan( data[8][index][freq_mask] ), data[7][index][freq_mask], label = resin_data.loc[i]["Label"], color = colours[i], alpha = alpha_scale, s = 25 )

                ax.set_rlim( 3, 170000 )
                ax.set_thetamin( 30 )
                ax.set_thetamax( 75 )
                ax.set_rscale( 'symlog' )

                r = np.arange( 0, 170000, 10 )
                theta = [np.pi / 4 for i in r]
                ax.plot( theta, r, "k--" )

            else:

                for i in samples_to_plot:

                    if specimens:

                        mask = np.where( sample_array == i )[0]

                        if not ip.shiny:

                            if all_specimens:

                                specimen_mask = [file_data[mask[j]][2] for j in range( len( mask ) )]

                            else:

                                specimen_mask = [file_data[mask[j]][2] for j in specimen_mask_by_index]

                        for j in mask:

                            if file_data[j][2] in specimen_mask:

                                if deriv0:

                                    if x == 0:

                                        if y == 4:

                                            data[y][j] = np.arctan( data[y][j] ) * 180 / np.pi

                                        xaxis_mask = np.where( (data[0] <= upper_bound) & (data[0] >= lower_bound) )[0]

                                        plt.plot( data[0][xaxis_mask], data[y][j][xaxis_mask], label = file_data[j][2], color = colours[i], linestyle = "None", marker = "o" )

                                        # plt.plot( data[0][xaxis_mask], data[1][j][xaxis_mask], label = "Storage Modulus", color = colours[i] )
                                        # plt.plot( data[0][xaxis_mask], data[2][j][xaxis_mask], label = "Loss Modulus", color = colours[i] )

                                        # plt.plot( data[0][freq_mask], data[0][freq_mask] ** m * np.exp( b ), "--", color = colours[i], linewidth = 2.5 )

                                        shiny_de.append( data[0][xaxis_mask].tolist() )
                                        shiny_de.append( data[y][j][xaxis_mask].tolist() )

                                        # m = (np.log( data[3][j][3] ) - np.log( data[3][j][1] )) / (np.log( data[0][3] ) - np.log( data[0][1] ))
                                        # b = np.log( data[3][j][2] / data[0][2] ** m )
                                        #
                                        data_extraction.append( data[0][xaxis_mask] )
                                        data_extraction.append( data[y][j][xaxis_mask] )
                                        # data_extraction.append( data[2][j][xaxis_mask] )

                                    else:

                                        if x == 4:

                                            data[x][j] = np.arctan( data[x][j] ) * 180 / np.pi

                                        if y == 4:

                                            data[y][j] = np.arctan( data[y][j] ) * 180 / np.pi

                                        xaxis_mask = np.where( (data[x][j] <= upper_bound) & (data[x][j] >= lower_bound) )[0]

                                        plt.plot( data[x][j][xaxis_mask], data[y][j][xaxis_mask], label = file_data[j][2], color = colours[i], linestyle = "None", marker = "o" )

                                        shiny_de.append( data[x][j][xaxis_mask].tolist() )
                                        shiny_de.append( data[y][j][xaxis_mask].tolist() )

                                if deriv1:

                                    freq_mask = np.where( (first_derivative_data[0] <= upper_bound) & (first_derivative_data[0] >= lower_bound) )[0]

                                    # Minus sign for log graph!

                                    plt.plot( first_derivative_data[0][freq_mask], -first_derivative_data[3][j][freq_mask], label = file_data[j][2], color = colours[i] )

                                if deriv2:

                                    freq_mask = np.where( (second_derivative_data[0] <= upper_bound) & (second_derivative_data[0] >= lower_bound) )[0]

                                    plt.plot( second_derivative_data[0][freq_mask], second_derivative_data[3][j][freq_mask], label = file_data[j][2], color = colours[i] )

                    if mean:

                        index = np.where( samples_present_array == i )[0][0]

                        if deriv0:

                            if x == 0:

                                if y == 8:

                                    data[y][index] = np.arctan( data[y][index] ) * 180 / np.pi

                                xaxis_mask = np.where( (data[0] <= upper_bound) & (data[0] >= lower_bound) )[0]

                                plt.plot( data[0][xaxis_mask], data[y][index][xaxis_mask], label = resin_data.loc[i]["Label"], color = colours[i] )

                                shiny_de.append( data[0][xaxis_mask].tolist() )
                                shiny_de.append( data[y][index][xaxis_mask].tolist() )

                            else:

                                if x == 8:

                                    data[x][index] = np.arctan( data[x][index] ) * 180 / np.pi

                                if y == 8:

                                    data[y][index] = np.arctan( data[y][index] ) * 180 / np.pi

                                xaxis_mask = np.where( (data[x][index] <= upper_bound) & (data[x][index] >= lower_bound) )[0]

                                plt.plot( data[x][index][xaxis_mask], data[y][index][xaxis_mask], label = resin_data.loc[i]["Label"], color = colours[i] )

                                shiny_de.append( data[x][index][xaxis_mask].tolist() )
                                shiny_de.append( data[y][index][xaxis_mask].tolist() )

                        if deriv1:

                            freq_mask = np.where( (first_derivative_data[0] <= upper_bound) & (first_derivative_data[0] >= lower_bound) )[0]

                            plt.plot( first_derivative_data[0][freq_mask], first_derivative_data[5][index][freq_mask], label = resin_data.loc[i]["Label"], color = colours[i] )

                        if deriv2:

                            freq_mask = np.where( (second_derivative_data[0] <= upper_bound) & (second_derivative_data[0] >= lower_bound) )[0]

                            plt.plot( second_derivative_data[0][freq_mask], second_derivative_data[7][index][freq_mask], label = resin_data.loc[i]["Label"], color = colours[i] )

            if ip.shiny:

                return shiny_de

            if log_graph:

                ax = plt.gca()
                ax.set_xscale( 'log' )
                ax.set_yscale( 'log' )

            # plt.legend( ncol = 3, bbox_to_anchor = ( 1.010, 0 ), loc = 'center', borderaxespad = 0 )
            leg = ax.get_legend()

            # for lh in leg.legendHandles:
            #
            #     lh.set_alpha(1)

            # plt.legend( ncol = 2, loc = 'upper right', borderaxespad = 0 )
            plt.legend()

            plt.xlabel( "Angular Frequency [rad/s]" )
            plt.ylabel( "Storage / Loss Modulus [Pa]" )

            plt.tight_layout()

            # For overall pipeline figure.

            # ax = plt.gca()
            # ax.get_legend().remove()
            # plt.xlabel( "" )
            # plt.ylabel( "" )
            # plt.tick_params( axis = 'x', which = 'both', bottom = False, top = False, labelbottom = False )
            # plt.tick_params( axis = 'y', which = 'both', left = False, right = False, labelleft = False )

            if savefig:

                plt.savefig( ip.output_directory + "Rheology/Plots/Plot.pdf" )

            else:

                plt.show()

            plt.close()

            if data_extraction_bool:

                array = data_extraction[0][:, np.newaxis]

                for i in range( 1, len( data_extraction ) ):

                    array = np.hstack( (array, data_extraction[i][:, np.newaxis]) )

                np.savetxt( ip.output_directory + "Plot_Coords/Unnamed.txt", array )

    return shiny_de



# Imports.

from . import Preprocessing
from . import Utilities as util

from .. import Global_Utilities as gu

# Main function definition.

def Global_Analysis_Main( ip ):

    if type( ip.datasets_to_read ) == int:

        ip.datasets_to_read = [ip.datasets_to_read]

    if ip.datasets_to_read == False:

        print( "Please select a dataset(s)" )

        return 0

    resin_data = gu.get_list_of_resins_data( ip.directory ) # Obtain the spreadsheet of data for the resins.

    if ip.read_files:

        features_df, std_of_features_df, rank_features_df = Preprocessing.read_files_and_preprocess( ip )

    if ip.plot_global_features:

        gu.plot_global_features( ip.output_directory + "Global/", features_df.to_numpy(), features_df.columns, [resin_data.loc[i]["Label"] for i in features_df.index] )

    if ip.scatterplot:

        return util.scatterplots( ip )

    if ip.correlation_heatmaps:

        spearman_rank_df = util.correlation_heatmap( rank_features_df, spearman = True )
        pearson_df = util.correlation_heatmap( features_df )

        gu.plot_df_heatmap( spearman_rank_df, savefig = True, filename = ip.output_directory + "Global/Correlations/Spearman.pdf" )
        gu.plot_df_heatmap( pearson_df, savefig = True, filename = ip.output_directory + "Global/Correlations/Pearson.pdf" )

    if ip.pca:

        return util.pca( ip.directory, ip.output_directory, ip.shiny, features_df, std_of_features_df )

    if ip.distance_to_virgin_analysis_based_on_pcas:

        util.distance_to_virgin_analysis_based_on_pcas( ip.output_directory, features_df )

    if ip.manual_ml:

        util.manual_ml( ip.directory, ip, features_df )

    if ip.pca_ml:

        util.pca_ml( ip.directory, ip, features_df )

    if ip.sandbox:

        util.sandbox( ip.directory, features_df, std_of_features_df )



from distinctipy import distinctipy
import numpy as np
import matplotlib.lines as mlines
import matplotlib.pyplot as plt
import matplotlib.cm as cm
from scipy.cluster.hierarchy import linkage
from scipy.cluster.hierarchy import dendrogram
from scipy.spatial.distance import squareform
from sklearn.decomposition import PCA
import pandas as pd
import re
from scipy.optimize import curve_fit
from adjustText import adjust_text
import math
import itertools
import sys
from pathlib import Path

sys.path.append( Path( __file__ ).absolute().parents[0].as_posix() + "/Global/" )

from .Global_Analysis import Utilities as util

def print_files_read( files_read, target ):

    print( str( files_read ) + " files have been read." )

    if files_read != target:

        print( "Warning: " + str( target ) + " files have not been read." )

def test_for_duplicates( file_data, data ):

    for i in range( len( file_data ) ):

        for j in range( i + 1, len( file_data ) ):

            if (data[i] == data[j]).all():

                print( file_data[i], file_data[j], "are the same!" )

def sort_raw_files_1( elem ):
    '''Files will have form ResinX, sorting according to X.'''

    pattern = re.compile( r"^Resin(\d+)" )

    return int( pattern.search( elem ).groups()[0] )

def sort_raw_files_2( elem ):
    '''Files will have form ResinX_Y_, sorting according to Y.'''

    pattern = re.compile( r"^Resin(\d+)_(\d+)_" )

    return int( pattern.search( elem ).groups()[1] )

def sort_raw_files_3( elem ):
    '''Files will have form ResinX_Y_, sorting according to X, then Y.'''

    pattern = re.compile( r"^Resin(\d+)_(\d+)_" )

    return [int( pattern.search( elem ).groups()[0] ), int( pattern.search( elem ).groups()[1] )]

def get_list_of_resins_data( directory, name_appendage = "" ):

    return pd.read_excel( directory + "List_of_Resins" + name_appendage + ".xlsx", header = 0, index_col = 0 )

def get_features_metadata( directory ):

    df = pd.read_excel( directory + "Features_Metadata.xlsx", header = 0, index_col = 0 )

    return df

def merge( file_data ):
    '''Merge resins that are technically the same material.'''

    # Resin40 becomes Resin5
    # Resin41 becomes Resin20
    # Resin1XX becomes ResinXX
    # Resin324 becomes Resin24

    for i in range( len( file_data ) ):

        if file_data[i][0] == 40:

            file_data[i][0] = 5

        if file_data[i][0] == 41:

            file_data[i][0] = 20

        if file_data[i][0] >= 101 and file_data[i][0] <= 200:

            file_data[i][0] -= 100

        if file_data[i][0] == 324:

            file_data[i][0] = 24

def extract_file_data_from_filenames( filenames, pattern_string ):
    """Returns file_data object of form [PCR Sample, PCR Specimen, Plot Label, Description]."""

    file_data = []

    pattern = re.compile( pattern_string )

    for f in filenames:

        sample = int( pattern.search( f ).groups()[0] )
        specimen = int( pattern.search( f ).groups()[1] )

        if sample == 0:

            file_data.append( [sample, specimen, "V{}.{}".format( 1, specimen ), ""] )

        elif sample == 16:

            file_data.append( [sample, specimen, "V{}.{}".format( 6, specimen ), ""] )

        elif sample == 17:

            file_data.append( [sample, specimen, "V{}.{}".format( 7, specimen ), ""] )

        elif sample == 19:

            file_data.append( [sample, specimen, "V{}.{}".format( 8, specimen ), ""] )

        else:

            file_data.append( [sample, specimen, "PCR{}.{}".format( sample, specimen ), ""] )

    return file_data

def sample_data_from_file_data( file_data ):
    '''Extract sample data from file data.'''

    sample = []

    for f in file_data:

        sample.append( f[0] )

    sample_array = np.array( sample )
    samples_present = sorted( list( set( sample ) ) )
    samples_present_array = np.array( samples_present )

    return sample, sample_array, samples_present, samples_present_array

def sample_mean( file_data, data ):
    '''Compute the mean for each sample.'''

    m = []

    sample, sample_array, samples_present, samples_present_array = sample_data_from_file_data( file_data )

    for i in samples_present:

        mean_array = []

        mask = np.where( sample_array == i )[0]

        for j in range( len( data[mask[0]] ) ):

            mean = 0

            for k in mask:

                mean += data[k][j]

            mean /= len( mask )

            mean_array.append( mean )

        m.append( np.array( mean_array ) )

    return m

def derivative( x, y, width = 1 ):
    '''Computes the derivative of a set of xy points.'''

    deriv = []

    for i in range( width, len( x ) - width ):

        if x[i + width] == x[i - width]:

            if i != width:

                deriv_at_point = deriv[i - width - 1]

            else:

                x = 1 / 0 # To retrun divide by zero error.

        else:

            deriv_at_point = (y[i + width] - y[i - width]) / (x[i + width] - x[i - width])

        deriv.append( deriv_at_point )

    return deriv

def integral_1( x, y, start, end, min_val = 0 ):
    '''Compute the integral of an xy curve between bounds. This is a true integral, perhaps with a subtraction of a rectangle with height the minimum value.'''

    integral = 0

    for i in range( start, end ):

        integral += abs( x[i] - x[i + 1] ) * 0.5 * (y[i] + y[i + 1] - 2 * min_val)

    return integral

def integral_2( x, y, start, end, min_val = 0 ):
    '''Compute the integral of an xy curve between bounds. This is the integral if the baseline is the line from the startpoint to the endpoint.'''

    integral = 0

    for i in range( start, end ):

        integral += abs( x[i] - x[i + 1] ) * 0.5 * (y[i] + y[i + 1] - 2 * min_val)

    integral -= abs( y[start] - y[end] ) * abs( x[start] - x[end] ) * 0.5

    return integral

def integral_3( y, width ):
    '''Compute the integral of an xy curve where the data points are at regular intervals (width).'''

    return (y[1: -1].sum() + 0.5 * (y[0] + y[-1])) * width

def array_with_column_titles_to_df( array, titles ):

    df = pd.DataFrame( array )
    df.columns = titles

    return df

def csv_to_df_to_array_and_column_titles( filename ):

    df = pd.read_csv( filename )

    df.drop( columns = [df.columns[0]], inplace = True )

    column_titles = df.columns.tolist()

    array_columns = []

    for i in range( len( df.columns ) ):

        array_columns.append( df.iloc[:, i].to_numpy() )

    array = array_columns[0][:, np.newaxis]

    for i in range( 1, len( array_columns ) ):

        array = np.hstack( (array, array_columns[i][:, np.newaxis]) )

    return column_titles, array

def remove_redundant_samples( sample_mask, samples_present ):

    to_pop = []

    for ind, i in enumerate( sample_mask ):

        if i not in samples_present:

            to_pop.append( ind )

    to_pop.reverse()

    for p in to_pop:

        sample_mask.pop( p )

    return sample_mask

def produce_mask( array_to_mask, desired_mask ):
    '''Generate a mask to apply given an array to mask and a desired mask.'''

    applied_mask = np.where( array_to_mask == desired_mask[0] )[0]

    for idx, k in enumerate( desired_mask ):

        if idx == 0:

            continue

        applied_mask = np.concatenate( [applied_mask, np.where( array_to_mask == k )[0]] )

    return applied_mask

def normalise_features( features, feature_weights = [] ):

    for i in range( features.shape[1] ):

        min = features[:, i].min()
        max = features[:, i].max()
        feature_range = max - min

        feature_weight = 1

        if feature_weights:

            feature_weight = feature_weights[i]

        features[:, i] = feature_weight * (features[:, i] - min) / feature_range

def combine_features( features, feature_names = [], feature_combinations = [] ):

    new_features = np.zeros( len( features ) )[:, np.newaxis]
    new_feature_names = []

    for i in feature_combinations:

        feature = np.zeros( len( features ) )

        feature_name = ""

        for j in i:

            feature += features[:, j]

            if not feature_name and feature_names:

                feature_name += feature_names[j]

        feature /= len( i )

        new_features = np.hstack( (new_features, np.array( feature )[:, np.newaxis]) )

        new_feature_names.append( feature_name )

    new_features = new_features[:, 1:]

    if not feature_combinations:

        return features, feature_names

    else:

        return new_features, new_feature_names

def extract_mean_features( features, sample_array, sample_mask ):

    mean_features = np.zeros( (len( sample_mask ), features.shape[1] ) )

    for ind, i in enumerate( sample_mask ):

        mask = np.where( np.array( sample_array ) == i )[0]

        for m in mask:

            for j in range( features.shape[1] ):

                mean_features[ind][j] += features[m][j]

        mean_features[ind] = mean_features[ind] / len( mask )

    return mean_features

def extract_mean_features_2( features, sample_array, sample_mask ):

    mean_features = np.zeros( (len( sample_mask ), features.shape[1] ) )

    for ind, i in enumerate( sample_mask ):

        mask = np.where( np.array( sample_array ) == i )[0]

        for m in mask:

            for j in range( features.shape[1] ):

                mean_features[ind][j] += features[m][j]

        for j in range( features.shape[1] ):

            mean_features[ind][j] -= features[mask, j].max()
            mean_features[ind][j] -= features[mask, j].min()

        mean_features[ind] = mean_features[ind] / (len( mask ) - 2)

    return mean_features

def extract_std_of_features( features, sample_array, sample_mask ):

    std_of_features = np.zeros( (len( sample_mask ), features.shape[1] ) )

    for ind, i in enumerate( sample_mask ):

        mask = np.where( np.array( sample_array ) == i )[0]

        if len( mask ) <= 1:

            for j in range( features.shape[1] ):

                std_of_features[ind][j] = 0

        else:

            for j in range( features.shape[1] ):

                std_of_features[ind][j] = np.std( features[mask, j], ddof = 1 ) / np.sqrt( len( mask ) )

    return std_of_features

def extract_range_of_features( features, sample_array, sample_mask ):

    range_of_features = np.zeros( (len( sample_mask ), features.shape[1] ) )

    for ind, i in enumerate( sample_mask ):

        mask = np.where( np.array( sample_array ) == i )[0]

        if len( mask ) <= 1:

            for j in range( features.shape[1] ):

                range_of_features[ind][j] = 0

        else:

            for j in range( features.shape[1] ):

                range_of_features[ind][j] = np.ptp( features[mask, j] )

    return range_of_features

def add_mean_feature( features, feature_names, sample_mask, new_feature, feature_name, weight = 1 ):

    feature = np.array( new_feature )[sample_mask]

    normalised_feature = (feature - feature.min()) / (feature.max() - feature.min()) * weight

    features = np.hstack( (features, np.array( normalised_feature )[:, np.newaxis]) )
    # features = np.hstack( (features, feature[:, np.newaxis]) )

    feature_names.append( feature_name )

    return features, feature_names

def compute_range_mean_std( data, mask ):
    '''Computes the range, median, mean and standard deviation of a set of curves.'''

    range_list = []
    high_point = []
    low_point = []
    median = []
    mean = []
    std = []

    for i in range( len( data[0] ) ):

        value = np.zeros( len( mask ) )

        for j, k in enumerate( mask ):

            value[j] = data[k][i]

        value = np.sort( value )

        range_list.append( value.max() - value.min() )
        high_point.append( value.max() )
        low_point.append( value.min() )

        if len( mask ) % 2 == 0:

            median.append( (value[int( len( mask ) / 2 )] + value[int( len( mask ) / 2 - 1 )]) / 2 )

        else:

            median.append( value[int( (len( mask ) - 1) / 2 )] )

        m = value.sum() / len( mask )
        mean.append( m )

        dev = 0

        for j in range( len( mask ) ):

            dev += (value[j] - m) * (value[j] - m)

        dev /= len( mask )

        dev = math.sqrt( dev )

        std.append( dev )

    return range_list, high_point, low_point, median, mean, std

def index_to_label( i ):

    if i == 24:

        return "V1"

    elif i < 0:

        return "V{}".format( 1 - i)

    elif i == 16:

        return "V1"

    elif i == 17:

        return "V2"

    elif i == 19:

        return "V3"

    else:

        return "PCR {}".format( i )

def plot_global_features( output_directory, features, feature_names, labels, subdirectory = "Features/", title = "Unnamed.pdf" ):

    features = np.transpose( features )

    fig, ax = plt.subplots()

    fig.set_size_inches( 5, 12 )

    im = ax.imshow( features, cmap = cm.plasma )
    plt.xticks( np.arange( 0, len( labels ), 1 ) )
    plt.yticks( np.arange( 0, len( feature_names ), 1 ) )
    plt.tick_params( left = False )

    ax.set_yticklabels( [n for n in feature_names], fontsize = 10 )
    # ax.set_yticklabels( labels = "" )
    ax.set_xticklabels( labels, rotation = 270, fontsize = 13 )

    linewidth = 6

    x, y = ([-0.5, 22.5, 22.5, -0.5, -0.5], [-0.5, -0.5, 32.5, 32.5, -0.5])
    line = mlines.Line2D( x, y, lw = linewidth, c = 'lime' )
    ax.add_line( line )

    x, y = ([-0.5, 22.5, 22.5, -0.5, -0.5], [32.5, 32.5, 37.5, 37.5, 32.5])
    line = mlines.Line2D( x, y, lw = linewidth, c = 'lime' )
    ax.add_line( line )

    x, y = ([-0.5, 22.5, 22.5, -0.5, -0.5], [37.5, 37.5, 43.5, 43.5, 37.5])
    line = mlines.Line2D( x, y, lw = linewidth, c = 'lime' )
    ax.add_line( line )

    x, y = ([-0.5, 22.5, 22.5, -0.5, -0.5], [43.5, 43.5, 48.5, 48.5, 43.5])
    line = mlines.Line2D( x, y, lw = linewidth, c = 'lime' )
    ax.add_line( line )

    x, y = ([-0.5, 22.5, 22.5, -0.5, -0.5], [48.5, 48.5, 52.5, 52.5, 48.5])
    line = mlines.Line2D( x, y, lw = linewidth, c = 'lime' )
    ax.add_line( line )

    x, y = ([-0.5, 22.5, 22.5, -0.5, -0.5], [52.5, 52.5, 55.5, 55.5, 52.5])
    line = mlines.Line2D( x, y, lw = linewidth, c = 'lime' )
    ax.add_line( line )

    # x, y = ([-0.5, 22.5, 22.5, -0.5, -0.5], [57.5, 57.5, 59.5, 59.5, 57.5])
    # line = mlines.Line2D( x, y, lw = linewidth, c = 'lime' )
    # ax.add_line( line )

    cbar = fig.colorbar( im, orientation = 'horizontal', location = 'top', shrink = 0.8, pad = 0.02 )
    cbar.ax.tick_params( labelsize = 13 )

    plt.tight_layout()

    plt.savefig( output_directory + subdirectory + title )

    plt.close()

def plot_global_features_transpose( output_directory, features, feature_names, labels, subdirectory = "Features/", title = "Unnamed.pdf" ):

    fig, ax = plt.subplots()

    fig.set_size_inches( 12, 6 )

    im = ax.imshow( features, cmap = cm.plasma )
    plt.xticks( np.arange( 0, len( feature_names ), 1 ) )
    plt.yticks( np.arange( 0, len( labels ), 1 ) )

    ax.set_xticklabels( [n for n in feature_names], rotation = 270, fontsize = 10 )
    ax.set_yticklabels( labels, fontsize = 10 )

    linewidth = 6

    y, x = ([-0.5, 22.5, 22.5, -0.5, -0.5], [-0.5, -0.5, 32.5, 32.5, -0.5])
    line = mlines.Line2D( x, y, lw = linewidth, c = 'lime' )
    ax.add_line( line )

    y, x = ([-0.5, 22.5, 22.5, -0.5, -0.5], [32.5, 32.5, 37.5, 37.5, 32.5])
    line = mlines.Line2D( x, y, lw = linewidth, c = 'lime' )
    ax.add_line( line )

    y, x = ([-0.5, 22.5, 22.5, -0.5, -0.5], [37.5, 37.5, 43.5, 43.5, 37.5])
    line = mlines.Line2D( x, y, lw = linewidth, c = 'lime' )
    ax.add_line( line )

    y, x = ([-0.5, 22.5, 22.5, -0.5, -0.5], [43.5, 43.5, 48.5, 48.5, 43.5])
    line = mlines.Line2D( x, y, lw = linewidth, c = 'lime' )
    ax.add_line( line )

    y, x = ([-0.5, 22.5, 22.5, -0.5, -0.5], [48.5, 48.5, 52.5, 52.5, 48.5])
    line = mlines.Line2D( x, y, lw = linewidth, c = 'lime' )
    ax.add_line( line )

    y, x = ([-0.5, 22.5, 22.5, -0.5, -0.5], [52.5, 52.5, 55.5, 55.5, 52.5])
    line = mlines.Line2D( x, y, lw = linewidth, c = 'lime' )
    ax.add_line( line )

    fig.colorbar( im, orientation = 'vertical', location = 'right', shrink = 0.75, pad = 0.02 )

    plt.tight_layout()

    plt.savefig( output_directory + subdirectory + title )

    plt.close()

def plot_boxplot( directory, features, feature_names, file_data, sample_mask, feature_name, xlabel = "", ylabel = "", filename = "", savefig = False ):

    resin_data = get_list_of_resins_data( directory )

    sample, sample_array, samples_present, samples_present_array = sample_data_from_file_data( file_data )

    sample_mask = remove_redundant_samples( sample_mask, samples_present )

    boxplot_list = []

    for i in sample_mask:

        specimen_mask = produce_mask( sample_array, [i] )

        features_2 = features[specimen_mask, feature_names.index( feature_name )]

        boxplot_list.append( features_2.tolist() )

    plt.boxplot( boxplot_list, whis = 10 )

    ax = plt.gca()

    ax.set_xticks( [y + 1 for y in range( len( boxplot_list ) )], labels = [resin_data.loc[i]["Label"] for i in sample_mask] )

    ax.set_xlabel( xlabel )
    ax.set_ylabel( ylabel )

    if savefig:

        plt.savefig( filename )

    else:

        plt.show()

    plt.close()

def plot_features( output_directory, features, feature_names, labels, specimen = False, subdirectory = "Features/", title = "Unnamed.pdf" ):

    fig, ax = plt.subplots()

    if specimen:

        fig.set_size_inches( 13, 18 )

    else:

        fig.set_size_inches( 8, 4 )

    im = ax.imshow( features, vmin = 0, vmax = 1, cmap = cm.plasma )
    plt.xticks( np.arange( 0, len( feature_names ), 1 ) )
    plt.yticks( np.arange( 0, len( labels ), 1 ) )

    if specimen:

        ax.set_xticklabels( [n for n in feature_names], rotation = 270, fontsize = 7 )
        ax.set_yticklabels( labels, fontsize = 7 )

    else:

        ax.set_xticklabels( [n for n in feature_names], rotation = 270, fontsize = 8 )
        ax.set_yticklabels( labels, fontsize = 7 )

    # [l.set_visible(False) for (i, l) in enumerate( ax.xaxis.get_ticklabels() ) if i % 2 != 0]

    cbar = fig.colorbar( im, orientation = 'vertical' )
    # cbar.ax.tick_params( labelsize = 9 )

    plt.tight_layout()
    ax.invert_yaxis()
    ax.invert_xaxis()

    plt.savefig( output_directory + subdirectory + title )

    plt.close()

def list_of_colours():
    '''Returns an object of 28 colours.'''

    colours = [(0, 0, 0)]

    colours.extend( distinctipy.get_colors( 130, pastel_factor = 0.2, rng = 2 ) )

    colours.extend( colours )
    colours.extend( colours )
    colours.extend( colours )

    return colours

def read_list_of_colours( directory ):

    df = pd.read_csv( directory + "List_of_Colours.csv" )

    return [(df.at[i, "r"], df.at[i, "g"], df.at[i, "b"]) for i in range( len( df.index ) )]

def write_list_of_colours_to_file( output_directory ):

    colours = list_of_colours()

    df = pd.DataFrame( colours )

    df.columns = ["r", "g", "b"]

    df.to_csv( output_directory + "List_of_Colours.csv" )

def plot_barchart_of_feature( features, labels, errorbars = False, std = [], colour = False, colour_mask = [], xlabel = "", ylabel = "", title = "", filename = "", savefig = False ):

    y_pos = np.arange( len( features ) )

    colours = list_of_colours()
    colours = [colours[i] for i in colour_mask]

    if colour:

        plt.bar( y_pos, features, align = 'center', alpha = 0.5, color = colours )

    else:

        plt.bar( y_pos, features, align = 'center', alpha = 0.5 )

    if errorbars:

        ax = plt.gca()

        for x, y, e, colour in zip( y_pos, features, std, colours ):

            ax.errorbar( x, y, e, capsize = 4, color = colour )

    plt.xticks( y_pos, labels, rotation = 90 )

    plt.xlabel( xlabel )
    plt.ylabel( ylabel )
    plt.title( title )

    plt.ylim( [55, 75] )

    plt.tight_layout()

    if savefig:

        plt.savefig( filename )

    else:

        plt.show()

    plt.close()

def distance_matrix_from_features( features ):

    distance_matrix = np.zeros( (features.shape[0], features.shape[0]) )

    for i in range( features.shape[0] ):

        for j in range( i + 1, features.shape[0] ):

            distance_matrix[i][j] = distance_matrix[j][i] = np.absolute( features[i] - features[j] ).sum()

    return distance_matrix

def array_with_column_titles_and_label_titles_to_df( array, column_titles, label_titles ):

    df = pd.DataFrame( array )
    df.columns = column_titles
    df = df.set_axis( label_titles, axis = "index" )

    return df

def subset_combinations_for_all_sizes_of_subsets( set_size ):

    set = range( set_size )

    combinations = []

    for L in range( set_size ):

        for subset in itertools.combinations( set, set_size - L ):

            combinations.append( list( subset ) )

    return combinations

def plot_boxes( ax, width, height, origin_x, origin_y ):
    '''Add boxes to a distance matrix around specimens of the same sample.'''

    x, y = ([origin_x, width + origin_x, width + origin_x, origin_x, origin_x], [origin_y, origin_y, height + origin_y, height + origin_y, origin_y])
    line = mlines.Line2D( x, y, lw = 1, c = 'w' )
    ax.add_line( line )

def plot_distance_matrix( output_directory, distance_matrix, labels, specimen = False, file_data = [], sample_mask = [], subdirectory = "Features/", title = "Unnamed.pdf" ):

    fig, ax = plt.subplots()

    if specimen:

        fig.set_size_inches( 20, 20 )

    else:

        fig.set_size_inches( 9, 9 )

    im = ax.imshow( distance_matrix, cmap = cm.plasma )

    # fig.colorbar( im, orientation = 'vertical' )

    plt.xticks( np.arange( 0, len( labels ), 1 ) )
    plt.yticks( np.arange( 0, len( labels ), 1 ) )

    if specimen:

        ax.set_xticklabels( labels, rotation = 270, fontsize = 6 )
        ax.set_yticklabels( labels, fontsize = 6 )

    else:

        ax.set_xticklabels( labels, rotation = 270, fontsize = 8 )
        ax.set_yticklabels( labels, fontsize = 8 )

    if not specimen:

        for i in range( len( labels ) ):

            for j in range( len( labels ) ):

                ax.text( j, i, "{:.0f}".format( distance_matrix[i][j] ), ha = "center", va = "center", color = "w" )

    if specimen:

        sample, sample_array, samples_present, samples_present_array = sample_data_from_file_data( file_data )

        num_specimens_of_each_sample = np.zeros( len( samples_present ) )

        for f in file_data:

            num_specimens_of_each_sample[np.argwhere( samples_present_array == f[0] )[0]] += 1

        cumulative_specimens_x = 0

        for i in range( len( sample_mask ) ):

            cumulative_specimens_y = 0

            for j in range( len( sample_mask ) ):

                plot_boxes( ax, num_specimens_of_each_sample[samples_present.index( sample_mask[i] )], num_specimens_of_each_sample[samples_present.index( sample_mask[j] )], -0.5 + cumulative_specimens_x, -0.5 + cumulative_specimens_y )

                cumulative_specimens_y += num_specimens_of_each_sample[samples_present.index( sample_mask[j] )]

            cumulative_specimens_x += num_specimens_of_each_sample[samples_present.index( sample_mask[i] )]

    plt.tight_layout()
    ax.invert_yaxis()

    plt.savefig( output_directory + subdirectory + title )

    plt.close()

def plot_dendrogram( output_directory, distance_matrix, labels, specimen = False, subdirectory = "Features/", title = "Unnamed.pdf" ):

    fig, ax = plt.subplots()

    if specimen:

        fig.set_size_inches( 18, 14 )

    else:

        fig.set_size_inches( 10, 8 )

    condensed_distance_matrix = squareform( distance_matrix )
    linkage_matrix = linkage( condensed_distance_matrix, "ward" )
    dendrogram( linkage_matrix, labels = labels, color_threshold = 0.65 * max( linkage_matrix[:,2] ), count_sort = True, leaf_rotation = 90 )

    plt.tight_layout()

    plt.savefig( output_directory + subdirectory + title )

    plt.close()

def csv_to_df_to_array_and_column_titles_and_label_titles( filename ):

    df = pd.read_csv( filename )

    label_titles = df[df.columns[0]].tolist()

    df.drop( columns = [df.columns[0]], inplace = True )

    column_titles = df.columns.tolist()

    array_columns = []

    for i in range( len( df.columns ) ):

        array_columns.append( df.iloc[:, i].to_numpy() )

    array = array_columns[0][:, np.newaxis]

    for i in range( 1, len( array_columns ) ):

        array = np.hstack( (array, array_columns[i][:, np.newaxis]) )

    return array, column_titles, label_titles

def func( x, a, b, c ):

    return a * np.exp( b * x ) + c

def exponential_curve_fit( feature_1, feature_2 ):

    feature_1 = np.array( feature_1 )
    feature_2 = np.array( feature_2 )

    xdata_max = feature_1.max()
    ydata_max = feature_2.max()

    xdata = feature_1 / xdata_max
    ydata = feature_2 / ydata_max

    temp = xdata.argsort()

    xdata = xdata[temp]
    ydata = ydata[temp]

    popt, pcov = curve_fit( func, xdata, ydata, maxfev = 800 )

    popt[0] = popt[0] * feature_2.max()
    popt[1] = popt[1] / feature_1.max()
    popt[2] = popt[2] * feature_2.max()

    return popt

def plot_scatterplot_of_two_features( directory, feature_1, feature_2, sample, labels, errorbars = False, std = [], line_of_best_fit = True, exponential_fit = False, xlog = False, ylog = False, title = "", xlabel = "", ylabel = "", annotate_style = 1, savefig = False, filename = "" ):

    colours = read_list_of_colours( directory )
    scatter_colours = [colours[sample[i]] for i in range( len( feature_1 ) )]
    # scatter_colours = sample

    if errorbars:

        ax = plt.gca()

        for x, y, xe, ye, colour in zip( feature_1, feature_2, std[0], std[1], scatter_colours ):

            ax.errorbar( x, y, xerr = xe, yerr = ye, fmt = "none", capsize = 4, ecolor = colour, zorder = 0 )

    sc = plt.scatter( feature_1, feature_2, color = scatter_colours, s = 100 )
    # sc = plt.scatter( feature_1, feature_2, edgecolor = "red", linewidths = 1, s = 120, c = scatter_colours )

    fig = plt.gcf()
    ax = plt.gca()

    # ax.plot( [np.min( [ax.get_xlim(), ax.get_ylim()] ), np.max( [ax.get_xlim(), ax.get_ylim()] )], [np.min( [ax.get_xlim(), ax.get_ylim()] ), np.max( [ax.get_xlim(), ax.get_ylim()] )], 'k-', alpha = 0.75, zorder = 0 )

    if line_of_best_fit:

        m, b = np.polyfit( feature_1, feature_2, 1 )

        plt.plot( feature_1, feature_1 * m + b )

    if exponential_fit:

        popt = exponential_curve_fit( feature_1, feature_2 )

        # funcx = np.linspace( 0.007, 1, 1000 )
        # funcx = np.linspace( -0.07, feature_1.max(), 1000 )
        funcx = np.linspace( feature_1.min(), feature_1.max(), 1000 )

        plt.plot( funcx, func( funcx, *popt ), 'r-', label = 'fit: a=%5.3f, b=%5.3f, c=%5.3f' % tuple( popt ) )

    plt.title( title )
    plt.xlabel( xlabel, fontsize = 13 )
    plt.ylabel( ylabel, fontsize = 13 )

    if annotate_style == 1:

        for i in range( len( feature_1 ) ):

            ax.annotate( labels[i], (feature_1[i] + 0.001, feature_2[i] + 0.001), fontsize = 13 )

    elif annotate_style == 2:

        TEXTS = []

        for i in range( len( feature_1 ) ):

            x = feature_1[i]
            y = feature_2[i]
            text = labels[i]

            TEXTS.append( ax.text( x, y, text, color = scatter_colours[i], fontsize = 14 ) )

        # 'expand_points' is a tuple with two multipliers by which to expand the bounding box of texts when repelling them from points.
        # 'arrowprops' receives a dictionary with all the properties we want for the arrows.

        # adjust_text( TEXTS, arrowprops = dict( arrowstyle = "->", lw = 2 ), expand_points = (3, 3), expand_text = (1, 1) )
        adjust_text( TEXTS, arrowprops = dict( arrowstyle = "->", lw = 2 ) )

    ax = plt.gca()

    if xlog == True:

        ax.set_xscale( 'log' )

    if ylog == True:

        ax.set_yscale( 'log' )

    # plt.xlim( [0, 1] )
    # plt.ylim( [0, 1] )

    plt.tight_layout()

    if savefig:

        plt.savefig( filename )

    else:

        plt.show()

    plt.close()

    # array = feature_1[:, np.newaxis]
    #
    # array = np.hstack( (array, feature_2[:, np.newaxis], np.array( std[0] )[:, np.newaxis], np.array( std[1] )[:, np.newaxis]) )
    #
    # np.savetxt( "/Users/philsmith/Documents/Postdoc/rHDPE_Data_Analysis/Output/Plot_Coords/Unnamed.txt", array )

    # array = 10 ** np.array( feature_1 )[:, np.newaxis]

    # array = np.hstack( (array, np.array( feature_2 )[:, np.newaxis]) )

    # array = np.hstack( (array, np.array( feature_2 )[:, np.newaxis], np.array( std[0] )[:, np.newaxis], np.array( std[1] )[:, np.newaxis]) )

    # array = np.pad( array, ((0, 976), (0, 0)) )

    # m, b = np.polyfit( feature_1, feature_2, 1 )
    #
    # array = np.hstack( (array, (np.array( feature_1 ) * m + b)[:, np.newaxis]) )

    # popt = exponential_curve_fit( feature_1, feature_2 )
    #
    # funcx = np.linspace( feature_1.min(), feature_1.max(), 1000 )
    #
    # array = np.hstack( (array, np.array( funcx )[:, np.newaxis], np.array( func( funcx, *popt ) )[:, np.newaxis]) )

    # np.savetxt( "/Users/philsmith/Documents/Postdoc/rHDPE_Data_Analysis/Output/Plot_Coords/Unnamed.txt", array )

def plot_scatterplot_of_two_features_pca( directory, feature_1, feature_2, coeff, coeff_names, sample, labels, errorbars = False, std = [], line_of_best_fit = True, exponential_fit = False, xlog = False, ylog = False, title = "", xlabel = "", ylabel = "", annotate_style = 1, savefig = False, filename = "" ):

    colours = read_list_of_colours( directory )
    scatter_colours = [colours[sample[i]] for i in range( len( feature_1 ) )]
    # scatter_colours = sample

    if errorbars:

        ax = plt.gca()

        for x, y, xe, ye, colour in zip( feature_1, feature_2, std[0], std[1], scatter_colours ):

            ax.errorbar( x, y, xerr = xe, yerr = ye, fmt = "none", capsize = 4, ecolor = colour, zorder = 0 )

    sc = plt.scatter( feature_1, feature_2, color = scatter_colours, s = 100 )

    fig = plt.gcf()
    ax = plt.gca()

    # ax.plot( [np.min( [ax.get_xlim(), ax.get_ylim()] ), np.max( [ax.get_xlim(), ax.get_ylim()] )], [np.min( [ax.get_xlim(), ax.get_ylim()] ), np.max( [ax.get_xlim(), ax.get_ylim()] )], 'k-', alpha = 0.75, zorder = 0 )

    if line_of_best_fit:

        m, b = np.polyfit( feature_1, feature_2, 1 )

        plt.plot( feature_1, feature_1 * m + b )

    if exponential_fit:

        xdata = feature_1
        ydata = feature_2

        temp = xdata.argsort()

        xdata = xdata[temp]
        ydata = ydata[temp]

        popt, pcov = curve_fit( func, xdata, ydata )

        funcx = np.linspace( 0.007, 1, 1000 )

        plt.plot( funcx, func( funcx, *popt ), 'r-', label = 'fit: a=%5.3f, b=%5.3f, c=%5.3f' % tuple( popt ) )

    n = coeff.shape[0]

    for i in range( n ):

        plt.arrow( 0, 0, coeff[i, 0], coeff[i, 1], color = 'r', alpha = 0.5 )

        plt.text( coeff[i, 0] * 1.15, coeff[i, 1] * 1.15, coeff_names[i], color = 'b', ha = 'center', va = 'center' )

    plt.title( title )
    plt.xlabel( xlabel, fontsize = 13 )
    plt.ylabel( ylabel, fontsize = 13 )

    if annotate_style == 1:

        for i in range( len( feature_1 ) ):

            ax.annotate( labels[i], (feature_1[i] + 0.001, feature_2[i] + 0.001), fontsize = 13 )

    elif annotate_style == 2:

        TEXTS = []

        for i in range( len( feature_1 ) ):

            x = feature_1[i]
            y = feature_2[i]
            text = labels[i]

            TEXTS.append( ax.text( x, y, text, color = scatter_colours[i], fontsize = 14 ) )

        # 'expand_points' is a tuple with two multipliers by which to expand the bounding box of texts when repelling them from points.
        # 'arrowprops' receives a dictionary with all the properties we want for the arrows.

        # adjust_text( TEXTS, arrowprops = dict( arrowstyle = "->", lw = 2 ), expand_points = (3, 3), expand_text = (1, 1) )
        adjust_text( TEXTS, arrowprops = dict( arrowstyle = "->", lw = 2 ) )

    ax = plt.gca()

    if xlog == True:

        ax.set_xscale( 'log' )

    if ylog == True:

        ax.set_yscale( 'log' )

    # plt.xlim( [0, 1] )
    # plt.ylim( [0, 1] )

    plt.tight_layout()

    if savefig:

        plt.savefig( filename )

    else:

        plt.show()

    plt.close()

    # array = feature_1[:, np.newaxis]
    #
    # array = np.hstack( (array, feature_2[:, np.newaxis], np.array( std[0] )[:, np.newaxis], np.array( std[1] )[:, np.newaxis]) )
    #
    # np.savetxt( "/Users/philsmith/Documents/Postdoc/rHDPE_Data_Analysis/Output/Plot_Coords/Unnamed.txt", array )

    # array = np.array( feature_1 )[:, np.newaxis]
    #
    # array = np.hstack( (array, feature_2[:, np.newaxis], np.array( std[0] )[:, np.newaxis], np.array( std[1] )[:, np.newaxis]) )
    #
    # for i in range( 2 ):
    #
    #     array_2 = np.pad( coeff[:, i], (0, len( array[:, 0] ) - len( coeff[:, i] )) )
    #
    #     array = np.hstack( (array, array_2[:, np.newaxis]) )
    #
    # np.savetxt( "/Users/philsmith/Documents/Postdoc/rHDPE_Data_Analysis/Output/Plot_Coords/Unnamed.txt", array )

def plot_scatterplot_of_three_features( directory, feature_1, feature_2, feature_3, samples_present, labels, title = "", xlabel = "", ylabel = "", zlabel = "", savefig = False, filename = "" ):

    colours = read_list_of_colours( directory )
    scatter_colours = [colours[samples_present[i]] for i in range( len( samples_present ) )]
    # scatter_colours = samples_present # Current method of plotting the actual resin colours.

    fig = plt.figure()
    ax = fig.add_subplot( projection = '3d' )

    ax.scatter3D( feature_1, feature_2, feature_3, s = 120, color = scatter_colours )
    # ax.scatter3D( feature_1, feature_2, feature_3, edgecolor = "red", linewidths = 1, s = 120, c = scatter_colours )

    for i in range( len( feature_1 ) ):

        ax.text( feature_1[i], feature_2[i], feature_3[i], labels[i], size = 8, zorder = 1,  color = 'k' )

    ax.set_title( title )
    ax.set_xlabel( xlabel )
    ax.set_ylabel( ylabel )
    ax.set_zlabel( zlabel )

    if savefig:

        plt.savefig( filename )

    else:

        plt.show()

    plt.close()

    array = np.array( feature_1 )[:, np.newaxis]

    array = np.hstack( (array, np.array( feature_2 )[:, np.newaxis]) )

    array = np.hstack( (array, np.array( feature_3 )[:, np.newaxis]) )

    np.savetxt( "/Users/philsmith/Documents/Postdoc/rHDPE_Data_Analysis/Output/Plot_Coords/Unnamed.txt", array )

def plot_scatterplot_of_two_features_with_hover_annotation( feature_1, feature_2, sample_mask, labels, xlabel = "", ylabel = "", title = "", savefig = False, filename = "" ):

    colours = list_of_colours()
    scatter_colours = [colours[sample_mask[i]] for i in range( len( feature_1 ) )]

    sc = plt.scatter( feature_1, feature_2, color = scatter_colours )

    lp = lambda i: plt.plot( [], color = colours[sample_mask[i]], label = labels[i], marker = "o" )[0]
    handles = [lp( i ) for i in range( len( sample_mask ) )]
    plt.legend( ncol = 2, bbox_to_anchor = ( 1.05, 1 ), loc = 'upper left', borderaxespad = 0 )

    fig, ax = plt.gcf(), plt.gca()

    annot = ax.annotate( "", xy = ( 0, 0 ), xytext = ( 20, 20 ), textcoords = "offset points", bbox = dict( boxstyle = "round", fc = "w" ), arrowprops = dict( arrowstyle = "->" ) )
    annot.set_visible( False )

    def update_annot( ind ):
        '''Update the annotation of a dot in the plot.'''

        pos = sc.get_offsets()[ind["ind"][0]]
        annot.xy = pos

        text = ""

        for i in range( len( ind["ind"] ) ):

            if i >= 1:

                text += ", "

            text += labels[ind["ind"][i]]

        annot.set_text( text )
        annot.get_bbox_patch().set_alpha( 0.4 )

    def hover( event ):
        '''Determine what happens when the mouse hovers over a point in the plot.'''

        vis = annot.get_visible()

        if event.inaxes == ax:

            cont, ind = sc.contains( event )

            if cont:

                update_annot( ind )
                annot.set_visible( True )
                fig.canvas.draw_idle()

            else:

                if vis:

                    annot.set_visible( False )
                    fig.canvas.draw_idle()

    fig.canvas.mpl_connect( "motion_notify_event", hover )

    plt.xlabel( xlabel )
    plt.ylabel( ylabel )
    plt.title( title )

    plt.tight_layout()

    if savefig:

        plt.savefig( filename )

    plt.show()

    plt.close()

def plot_scatterplot_of_two_specimen_features_with_hover_annotation( feature_1, feature_2, file_data, xlabel = "", ylabel = "", title = "", savefig = False, filename = "" ):

    sample, sample_array, samples_present, samples_present_array = sample_data_from_file_data( file_data )

    colours = list_of_colours()
    scatter_colours = [colours[file_data[i][0]] for i in range( len( feature_1 ) )]

    sc = plt.scatter( feature_1, feature_2, color = scatter_colours )

    lp = lambda i: plt.plot( [], color = colours[samples_present[i]], label = index_to_label( samples_present[i] ), marker = "o" )[0]
    handles = [lp( i ) for i in range( len( samples_present ) )]
    plt.legend( handles = handles )
    plt.legend( ncol = 2, bbox_to_anchor = ( 1.05, 1 ), loc = 'upper left', borderaxespad = 0 )

    fig, ax = plt.gcf(), plt.gca()

    annot = ax.annotate( "", xy = ( 0, 0 ), xytext = ( 20, 20 ), textcoords = "offset points", bbox = dict( boxstyle = "round", fc = "w" ), arrowprops = dict( arrowstyle = "->" ) )
    annot.set_visible( False )

    def update_annot( ind ):
        '''Update the annotation of a dot in the plot.'''

        pos = sc.get_offsets()[ind["ind"][0]]
        annot.xy = pos

        text = ""

        for i in range( len( ind["ind"] ) ):

            if i >= 1:

                text += ", "

            text += file_data[ind["ind"][i]][2]

        annot.set_text( text )
        annot.get_bbox_patch().set_alpha( 0.4 )

    def hover( event ):
        '''Determine what happens when the mouse hovers over a point in the plot.'''

        vis = annot.get_visible()

        if event.inaxes == ax:

            cont, ind = sc.contains( event )

            if cont:

                update_annot( ind )
                annot.set_visible( True )
                fig.canvas.draw_idle()

            else:

                if vis:

                    annot.set_visible( False )
                    fig.canvas.draw_idle()

    fig.canvas.mpl_connect( "motion_notify_event", hover )

    plt.xlabel( xlabel )
    plt.ylabel( ylabel )
    plt.title( title )

    plt.tight_layout()

    if savefig:

        plt.savefig( filename )

    plt.show()

    plt.close()

def plot_df_heatmap( df, savefig = False, filename = "" ):

    fig, ax = plt.subplots()

    fig.set_size_inches( 20, 20 )

    im = ax.imshow( df, cmap = cm.plasma )

    plt.xticks( np.arange( 0, len( df.columns ), 1 ) )
    plt.yticks( np.arange( 0, len( df.index ), 1 ) )

    ax.set_xticklabels( df.columns, rotation = 270, fontsize = 6 )
    ax.set_yticklabels( df.index, fontsize = 6 )

    plt.tight_layout()
    ax.invert_yaxis()

    # for i in range( len( df.columns ) ):
    #
    #         for j in range( len( df.index ) ):
    #
    #             ax.text( i, j, "{:.1f}".format( df.iloc[j, i] ), ha = "center", va = "center", color = "w" )

    if savefig:

        plt.savefig( filename )

    else:

        plt.show()

    plt.close()

def pca_analysis( pca, features_df, plot_pc_heatmaps = False ):

    print( "Full explained ratio: ", pca.explained_variance_ratio_, "\n" )

    components_to_keep = [i for i in range( len( pca.explained_variance_ratio_ ) ) if pca.explained_variance_ratio_[i] > 0.1]

    num_components = len( components_to_keep )

    most_important_features = []

    for i in range( num_components ):

        most_important_features.append( np.argsort( -np.abs( pca.components_[i] ) ) )

        most_important_features[i] = [j for j in most_important_features[i] if abs( pca.components_[i][j] ) > 0.05] # 0.05

        print( "Key features for PC ", i + 1, ":\n" )

        for j in range( len( most_important_features[i] ) ):

            print( pca.feature_names_in_[most_important_features[i][j]], pca.components_[i][most_important_features[i][j]] )

        print( "\n" )

        if plot_pc_heatmaps:

            pc_df = features_df[pca.feature_names_in_[most_important_features[i]]]

            plot_df_heatmap( pc_df )

    nums = np.arange( len( pca.explained_variance_ratio_ ) + 1 )
    var_ratio = [np.sum( pca.explained_variance_ratio_[:i] ) for i in range( len( pca.explained_variance_ratio_ ) + 1 )]

    # plt.figure( figsize = ( 4, 2 ), dpi = 150 )
    # plt.grid()
    # plt.plot( nums, var_ratio, marker = 'o' )
    # plt.xlabel( 'n_components' )
    # plt.ylabel( 'Explained variance ratio' )
    # plt.title( 'n_components vs. Explained Variance Ratio' )
    # plt.show()
    # plt.close()

    # print( var_ratio )

def perform_pca( directory, features_df, sample_mask, std_error = False, std_of_features_df = [], num_components = 3, filename = "Unnamed.pdf", analysis_of_pca = True ):

    resin_data = get_list_of_resins_data( directory ) # Obtain the spreadsheet of data for the resins.

    pca = PCA( n_components = num_components )
    pca_ft = pca.fit_transform( features_df )

    if analysis_of_pca:

        pca_analysis( pca, features_df )

    pca_ft_df = array_with_column_titles_and_label_titles_to_df( pca_ft, ["PC{}".format( i ) for i in range( num_components )], sample_mask )

    if std_error:

        std = [[] for i in range( num_components )]

        for k in range( num_components ):

            for i in range( len( std_of_features_df.iloc[:, 0] ) ):

                s = 0

                for j in range( len( pca.components_[k] ) ):

                    s += abs( pca.components_[k][j] ) * std_of_features_df.iloc[i, j]

                std[k].append( s )

        if num_components == 2:

            plot_scatterplot_of_two_features_pca( directory, pca_ft_df[pca_ft_df.columns[0]].to_numpy(), pca_ft_df[pca_ft_df.columns[1]].to_numpy(), np.transpose( pca.components_[0:2, :] ), pca.feature_names_in_, pca_ft_df.index, [resin_data.loc[i]["Label"] for i in pca_ft_df.index], errorbars = True, std = [std[0], std[1]], line_of_best_fit = False, xlabel = "First Principal Component", ylabel = "Second Principal Component", annotate_style = 2, savefig = True, filename = filename )

        elif num_components == 3:

            plot_scatterplot_of_three_features( directory, pca_ft_df[pca_ft_df.columns[0]].to_numpy(), pca_ft_df[pca_ft_df.columns[1]].to_numpy(), pca_ft_df[pca_ft_df.columns[2]].to_numpy(), pca_ft_df.index, [resin_data.loc[i]["Label"] for i in pca_ft_df.index], xlabel = "First Principal Component", ylabel = "Second Principal Component", zlabel = "Third Principal Component" )

    else:

        if num_components == 2:

            plot_scatterplot_of_two_features_pca( directory, pca_ft_df[pca_ft_df.columns[0]].to_numpy(), pca_ft_df[pca_ft_df.columns[1]].to_numpy(), np.transpose( pca.components_[0:2, :] ), pca.feature_names_in_, pca_ft_df.index, [resin_data.loc[i]["Label"] for i in pca_ft_df.index], line_of_best_fit = False, xlabel = "First Principal Component", ylabel = "Second Principal Component", annotate_style = 2, savefig = True, filename = filename )

        elif num_components == 3:

            plot_scatterplot_of_three_features( directory, pca_ft_df[pca_ft_df.columns[0]].to_numpy(), pca_ft_df[pca_ft_df.columns[1]].to_numpy(), pca_ft_df[pca_ft_df.columns[2]].to_numpy(), pca_ft_df.index, [resin_data.loc[i]["Label"] for i in pca_ft_df.index], xlabel = "First Principal Component", ylabel = "Second Principal Component", zlabel = "Third Principal Component" )

    if std_error:

        return pca_ft_df, [std[0], std[1]], np.transpose( pca.components_[0:2, :] ), pca.feature_names_in_

    else:

        return pca_ft_df, [], np.transpose( pca.components_[0:2, :] ), pca.feature_names_in_

def relabel_index_and_delete_first_column( df ):

    indices = df[df.columns[0]].tolist()
    df.set_axis( indices, axis = 0, inplace = True )
    df.drop( columns = df.columns[0], inplace = True )

    return indices

def plot_kmeans_plus_pca( features, kmeans, sample_mask, labels, title = "", xlabel = "", ylabel = "", savefig = False, filename = "" ):

    # Step size of the mesh. Decrease to increase the quality of the VQ.

    h = 0.005  # point in the mesh [x_min, x_max]x[y_min, y_max].

    # Plot the decision boundary. For that, we will assign a color to each

    x_min, x_max = features[:, 0].min() - 0.1, features[:, 0].max() + 0.1
    y_min, y_max = features[:, 1].min() - 0.1, features[:, 1].max() + 0.1
    xx, yy = np.meshgrid( np.arange( x_min, x_max, h ), np.arange( y_min, y_max, h ) )

    # Obtain labels for each point in mesh. Use last trained model.

    Z = kmeans.predict( np.c_[xx.ravel(), yy.ravel()] )

    # Put the result into a color plot

    Z = Z.reshape( xx.shape )
    plt.figure()
    plt.gcf().set_size_inches( 15, 15 )
    plt.clf()
    plt.imshow( Z, interpolation = "nearest", extent = ( xx.min(), xx.max(), yy.min(), yy.max() ), cmap = plt.cm.Paired, aspect = "auto", origin = "lower" )

    colours = list_of_colours()
    scatter_colours = [colours[sample_mask[i]] for i in range( len( sample_mask ) )]

    sc = plt.scatter( features[:, 0], features[:, 1], color = scatter_colours, s = 300 )

    fig = plt.gcf()
    ax = plt.gca()

    for i in range( len( sample_mask ) ):

        ax.annotate( labels[i], (features[:, 0][i] + 0.01, features[:, 1][i] + 0.01), fontsize = 30 )

    # Plot the centroids as a white X

    centroids = kmeans.cluster_centers_
    # plt.scatter( centroids[:, 0], centroids[:, 1], marker = "o", color = "w", zorder = 10 )

    plt.title( title )
    plt.xlabel( xlabel, fontsize = 30 )
    plt.ylabel( ylabel, fontsize = 30 )

    if savefig:

        plt.savefig( filename )

    else:

        plt.show()

    plt.close()



# Imports.

import os
import glob
import re
import numpy as np
import pandas as pd

from .. import Global_Utilities as gu

# Function definitions.

def read_raw_data_file_1( filename, f, resin_data, file_data, data ):

    pattern = re.compile( r"^Resin(\d+)_(\d+)_" )

    resin = int( pattern.search( f ).groups()[0] )

    specimen = int( pattern.search( f ).groups()[1] )

    with open( filename, 'r' ) as file:

        x, y = [], []

        lines = file.readlines()

        for line in lines:

            a_list = line.split()

            if a_list:

                map_object = map( float, a_list )
                list_of_floats = list( map_object )

                if list_of_floats[0] <= 3996.26214 and list_of_floats[0] >= 599.82506:

                    x.append( list_of_floats[0] )
                    y.append( list_of_floats[1] )

            else:

                break

    data[0].append( np.array( x ) )
    data[1].append( np.array( y ) )

    file_data.append( [resin, specimen, resin_data.loc[resin]["Label"] + ".{}".format( specimen ), ""] )

def read_raw_data_file_2( filename, f, resin_data, file_data, data ):

    pattern = re.compile( r"^Resin(\d+)_(\d+)_" )

    resin = int( pattern.search( f ).groups()[0] )

    specimen = int( pattern.search( f ).groups()[1] )

    with open( filename, 'r' ) as file:

        x, y = [], []

        lines = file.read().splitlines()

        for line in lines:

            a_list = line.split( "," )

            if a_list:

                map_object = map( float, a_list )
                list_of_floats = list( map_object )

                if list_of_floats[0] <= 3996.26214 and list_of_floats[0] >= 599.82506:

                    x.append( list_of_floats[0] )
                    y.append( list_of_floats[1] )

            else:

                break

    data[0].append( np.array( x ) )
    data[1].append( np.array( y ) )

    file_data.append( [resin, specimen, resin_data.loc[resin]["Label"] + ".{}".format( specimen ), ""] )

def extract_raw_data( directory, data_directory ):
    '''Extract the raw data from the files.'''

    resin_data = gu.get_list_of_resins_data( directory ) # Obtain the spreadsheet of data for the resins.

    resins = sorted( [os.path.basename( path ) for path in glob.glob( data_directory + "*" )], key = gu.sort_raw_files_1 )

    file_data, data = [], [[], []]

    pattern = re.compile( r"^Resin(\d+)" )

    for r in resins:

        filenames = sorted( [os.path.basename( path ) for path in glob.glob( data_directory + r + "/*" )], key = gu.sort_raw_files_2 )

        resin = int( pattern.search( r ).groups()[0] )

        for f in filenames:

            if (resin >= 40 and resin <= 70) or (resin >= 601 and resin <= 605) or (resin >= 701 and resin <= 706):

                read_raw_data_file_2( data_directory + r + "/" + f, f, resin_data, file_data, data )

            else:

                read_raw_data_file_1( data_directory + r + "/" + f, f, resin_data, file_data, data )

    return file_data, data

def normalise( y ):
    '''Normalise the data.'''

    for i in range( len( y ) ):

        y[i] = y[i] / y[i].max()

def normalise_2( data ):
    '''Normalise the data.'''

    for i in range( len( data[1] ) ):

        data[1][i] = (data[1][i] - data[1][i].min()) / (data[1][i].max() - data[1][i].min())

    range_mask_1 = np.where( (data[0] >= 712) & (data[0] < 726) )
    range_mask_2 = np.where( (data[0] >= 1455) & (data[0] < 1468) )

    range_mask_3 = np.where( (data[0] >= 580) & (data[0] < 720) )
    range_mask_4 = np.where( (data[0] >= 720) & (data[0] < 1468) )
    range_mask_5 = np.where( (data[0] >= 1468) & (data[0] < 2915) )

    for i in range( len( data[1] ) ):

        value_1 = data[1][i][range_mask_1].max()
        value_2 = data[1][i][range_mask_2].max()

        scale_1 = value_1 / 0.09
        scale_2 = value_2 / 0.15

        data[1][i][range_mask_3] = data[1][i][range_mask_3] / np.linspace( scale_1, 1, len( range_mask_3[0] ) )
        data[1][i][range_mask_4] = data[1][i][range_mask_4] / np.linspace( scale_2, scale_1, len( range_mask_4[0] ) )
        data[1][i][range_mask_5] = data[1][i][range_mask_5] / np.linspace( 1, scale_2, len( range_mask_5[0] ) )

def standardise_data( data ):
    '''Standardise data.'''

    standard_x_list = np.linspace( 600, 3996, 1762 ).tolist()
    standard_x_list.reverse()

    for i in range( len( data[0] ) ):

        mask = []

        array = np.array( data[0][i] )

        for j in range( len( standard_x_list ) ):

            interval_mask_upper = np.where( array >= standard_x_list[j] )[0]
            interval_mask_lower = np.where( array <= standard_x_list[j] )[0]

            if len( interval_mask_upper ) == 0:

                mask.append( interval_mask_lower[0] )

            elif len( interval_mask_lower ) == 0:

                mask.append( interval_mask_upper[len( interval_mask_upper ) - 1] )

            elif abs( data[0][i][interval_mask_upper[len( interval_mask_upper ) - 1]] - standard_x_list[j] ) > abs( data[0][i][interval_mask_lower[0]] - standard_x_list[j] ):

                mask.append( interval_mask_lower[0] )

            else:

                mask.append( interval_mask_upper[len( interval_mask_upper ) - 1] )

        data[1][i] = np.array( data[1][i] )[mask]

        print( i )

    data[0] = np.array( standard_x_list )

    normalise( data[1] )

    # normalise_2( data )

def add_description_to_file_data( file_data ):
    '''Add descriptions in the form of letters to each specimen.'''

    # Add m for data produced in Manchester (by Thomas Franklin).
    # Add f for probably faulty experimental data.
    # Add r for repetitions.
    # Add p for virgin specimen that (erroneously) contains significant PP.
    # Add n for noisy.
    # Add a for additive data.
    # Add e for ebm trial sprectra.
    # Add u for Unilever FTIRs.
    # Add b for blends.
    # Add d for blends that stand a little bit out from the pattern.
    # Add s for blends resins that seem somewhat separate from other specimens.
    # Add t for resins that came from Arpan to test the model.
    # Add v for resins that came from Unilever to test the model.

    specimens = {40:[0, 1, 2], 41:[0, 1, 2, 3, 4, 5]}

    for f in file_data:

        if f[0] in specimens.keys():

            if f[1] in specimens[f[0]]:

                f[3] = f[3] + "m"

    specimens = {40:[2], 406:[5]}

    for f in file_data:

        if f[0] in specimens.keys():

            if f[1] in specimens[f[0]]:

                f[3] = f[3] + "n"

    specimens = {4:[2, 4], 3:[0]}

    for f in file_data:

        if f[0] in specimens.keys():

            if f[1] in specimens[f[0]]:

                f[3] = f[3] + "f"

    specimens = {4:[5], 10:[8]}

    for f in file_data:

        if f[0] in specimens.keys():

            if f[1] in specimens[f[0]]:

                f[3] = f[3] + "r"

    specimens = {} #{19:[8]}

    for f in file_data:

        if f[0] in specimens.keys():

            if f[1] in specimens[f[0]]:

                f[3] = f[3] + "p"

    specimens = {41:[3, 4, 5]}

    for f in file_data:

        if f[0] in specimens.keys():

            if f[1] in specimens[f[0]]:

                f[3] = f[3] + "a"

    for f in file_data:

        if f[0] >= 50 and f[0] <= 100:

            f[3] = f[3] + "a"

    for f in file_data:

        if f[0] >= 301 and f[0] <= 400:

            f[3] = f[3] + "e"

    for f in file_data:

        if f[0] >= 401 and f[0] <= 499:

            f[3] = f[3] + "u"

    for f in file_data:

        if f[0] >= 500 and f[0] <= 599:

            f[3] = f[3] + "b"

    specimens = {502:[9], 506:[0, 1, 9], 508:[4, 7], 510:[9], 512:[3, 4, 9], 514:[0, 2], 516:[2]}

    for f in file_data:

        if f[0] in specimens.keys():

            if f[1] in specimens[f[0]]:

                f[3] = f[3] + "d"

    for f in file_data:

        if f[0] >= 600 and f[0] <= 699:

            f[3] = f[3] + "t"

    for f in file_data:

        if f[0] >= 700 and f[0] <= 799:

            f[3] = f[3] + "v"

def read_files_and_preprocess( directory, data_directory, merge_groups ):
    '''Read files and preprocess data.'''

    file_data, data = extract_raw_data( directory, data_directory )

    standardise_data( data )

    add_description_to_file_data( file_data )

    if merge_groups:

        gu.merge( file_data )

    return file_data, data

def read_shiny_file( directory, filepath, filename, name_appendage = "" ):

    resin_data = gu.get_list_of_resins_data( directory, name_appendage ) # Obtain the spreadsheet of data for the resins.

    pattern = re.compile( r"^Resin(\d+)_(\d+)_" )

    resin = int( pattern.search( filename ).groups()[0] )

    specimen = int( pattern.search( filename ).groups()[1] )

    with open( filepath, 'r' ) as file:

        x, y = [], []

        lines = file.readlines()

        for line in lines:

            a_list = line.split()

            if a_list:

                map_object = map( float, a_list )
                list_of_floats = list( map_object )

                if list_of_floats[0] <= 3996.26214 and list_of_floats[0] >= 599.82506:

                    x.append( list_of_floats[0] )
                    y.append( list_of_floats[1] )

            else:

                break

    file_data, data = [], [[], []]

    data[0].append( np.array( x ) )
    data[1].append( np.array( y ) )

    standardise_data( data )

    data[1][0] = data[1][0]

    file_data.append( [resin, specimen, resin_data.loc[resin]["Label"] + ".{}".format( specimen ), ""] )

    return file_data, data[1]

def write_csv( output_directory, file_data, data, name_appendage = "" ):
    '''Write read and preprocessed data to a .csv file.'''

    array = data[0][:, np.newaxis]

    for i in range( len( data[1] ) ):

        array = np.hstack( (array, data[1][i][:, np.newaxis]) )

    np.savetxt( output_directory + "FTIR/Condensed_Data/FTIR_data" + name_appendage + ".csv", array, delimiter = ",", fmt = "%.5f" )

    array = np.array( file_data )

    np.savetxt( output_directory + "FTIR/Condensed_Data/file_data" + name_appendage + ".csv", array, delimiter = ",", fmt = "%s" )

def read_csv( directory, output_directory, merge_groups, name_appendage = "" ):
    '''Read the preprocessed .csv file.'''

    resin_data = gu.get_list_of_resins_data( directory, name_appendage ) # Obtain the spreadsheet of data for the resins.

    file_data = []

    df = pd.read_csv( output_directory + "FTIR/Condensed_Data/file_data" + name_appendage + ".csv", sep = ",", header = None )

    for i in range( len( df.index ) ):

        resin = int( df.iloc[i, 0] )
        specimen = int( df.iloc[i, 1] )

        file_data.append( [resin, specimen, resin_data.loc[resin]["Label"] + ".{}".format( specimen ), ""] )

    data = []

    df = pd.read_csv( output_directory + "FTIR/Condensed_Data/FTIR_data" + name_appendage + ".csv", sep = ",", header = None )

    data.append( df.iloc[:, 0].to_numpy( dtype = np.float64 ) )

    y = []

    for i in range( 1, len( df.columns ) ):

        y.append( df.iloc[:, i].to_numpy( dtype = np.float64 ) )

    data.append( y )

    add_description_to_file_data( file_data )

    if merge_groups:

        gu.merge( file_data )

    return file_data, data

def remove_files( file_data, data, descriptors_to_remove = "" ):
    '''Remove files not needed/wanted for analysis by searching for letters in file descriptions.'''

    files_to_remove = []

    for i in range( len( file_data ) ):

        s = file_data[i][3]

        for l in descriptors_to_remove:

            if s.find( l ) > -0.5:

                files_to_remove.append( i )

                break

    files_to_remove.reverse()

    for r in files_to_remove:

        file_data.pop( r )
        data[1].pop( r )

    return file_data, data

def compute_mean( output_directory, file_data, data, name_appendage = "" ):
    '''Compute the mean data for each resin.'''

    m = gu.sample_mean( file_data, data[1] )

    array = m[0][:, np.newaxis]

    for i in range( 1, len( m ) ):

        array = np.hstack( (array, m[i][:, np.newaxis]) )

    np.savetxt( output_directory + "FTIR/Condensed_Data/Means" + name_appendage + ".csv", array, delimiter = ",", fmt = "%.5f" )

def read_mean( output_directory, data, name_appendage = "" ):
    '''Read the computed means for each resin from a file.'''

    m = []

    df = pd.read_csv( output_directory + "FTIR/Condensed_Data/Means" + name_appendage + ".csv", sep = ",", header = None )

    for i in range( len( df.columns ) ):

        m.append( df.iloc[:, i].to_numpy( dtype = np.float64 ) )

    if len( data ) > 2: # A previous list of means is already appended.

        data[2] = m

    else:

        data.append( m )

    return data



# Imports.

import numpy as np

from .. import Global_Utilities as gu

# Function definitions.

def compute_derivatives( data, width = 1 ):
    '''Compute the derivatives.'''

    derivative_data = [[], data[1][width: -width], [], [], [], []]

    for i in range( len( data[3] ) ):

        derivative_data[3].append( np.array( gu.derivative( data[1], data[3][i], width ) ) )

    for i in range( len( data[5] ) ):

        derivative_data[5].append( np.array( gu.derivative( data[1], data[5][i], width ) ) )

    return derivative_data

def extract_TGA_features( output_directory, file_data, data, first_derivative_data, second_derivative_data ):

    # Residue

    feature_1 = []

    for i in range( len( file_data ) ):

        feature_1.append( data[3][i][len( data[3][i] ) - 1] )

    features = np.array( feature_1 )[:, np.newaxis]

    feature_names = ["TGA_Residue"]

    # Mean derivative at a selection of 20 degree ranges

    for j in range( 28 ):

        feature_1 = []

        lower_bound = 20 * (j + 2)
        upper_bound = 20 * (j + 3)

        mask = np.where( (first_derivative_data[1] >= lower_bound) & (first_derivative_data[1] <= upper_bound) )[0]

        for i in range( len( file_data ) ):

            feature_1.append( -np.mean( first_derivative_data[3][i][mask] ) )

        features = np.hstack( (features, np.array( feature_1 )[:, np.newaxis]) )

        feature_names.append( "TGA_Temp_{}-{}".format( lower_bound, upper_bound ) )

    df = gu.array_with_column_titles_to_df( features, feature_names )

    df.to_csv( output_directory + "TGA/Features/Features.csv" )

def read_and_analyse_TGA_features( ip, file_data ):

    plot_specimen_bars = False
    plot_mean_bars = False
    plot_specimen_features = True
    plot_mean_features = True
    plot_specimen_distance_matrix = True
    plot_mean_distance_matrix = True
    plot_specimen_dendrogram = True
    plot_mean_dendrogram = True

    resin_data = gu.get_list_of_resins_data( ip.directory )

    feature_names, features = gu.csv_to_df_to_array_and_column_titles( ip.output_directory + "TGA/Features/Features.csv" )

    sample, sample_array, samples_present, samples_present_array = gu.sample_data_from_file_data( file_data )

    if not ip.sample_mask:

        ip.sample_mask = [11, 14, 10, 4, 13, 21, 23, 18, 22, 20, 2, 3, 17, 16, 19, 1, 15, 12, 6, 5, 7, 9, 8, 24]

    sample_mask = ip.sample_mask

    sample_mask = gu.remove_redundant_samples( sample_mask, samples_present )

    if not ip.feature_selection:

        # ip.feature_selection = [i for i in range( len( feature_names ) )]
        ip.feature_selection = [0, 7, 14, 18, 22, 23]

    feature_selection = ip.feature_selection

    features = features[:, feature_selection]
    feature_names = list( np.array( feature_names )[feature_selection] )
    # feature_names = ["Residue", "Feature 5", "Feature 4", "Feature 3", "Feature 2", "Feature 1"]

    mean_features_unnormalised = gu.extract_mean_features( features, sample_array, samples_present )
    mean_feature_names = feature_names.copy()

    mean_features_unnormalised_plus_sample_mask = np.hstack( (samples_present_array[:, np.newaxis], mean_features_unnormalised) )

    df = gu.array_with_column_titles_to_df( mean_features_unnormalised_plus_sample_mask, ["sample"] + mean_feature_names )

    df.to_csv( ip.output_directory + "TGA/Features/Mean_Features_Unnormalised.csv" )

    std_of_features = gu.extract_std_of_features( features, sample_array, samples_present )

    std_of_features_plus_sample_mask = np.hstack( (samples_present_array[:, np.newaxis], std_of_features) )

    df = gu.array_with_column_titles_to_df( std_of_features_plus_sample_mask, ["sample"] + mean_feature_names )

    df.to_csv( ip.output_directory + "TGA/Features/Std_of_Features_Unnormalised.csv" )

    specimen_mask = gu.produce_mask( sample_array, sample_mask )

    features = features[specimen_mask, :]
    file_data_mask = np.array( file_data )[specimen_mask] # Orders file data according to specimen mask.
    file_data_mask = [[int( f[0] ), int( f[1] ), f[2], f[3]] for f in file_data_mask] # Converts file data so that f[0], etc. are integers and not np.str.
    sample_array = sample_array[specimen_mask]

    feature_weights = [1 for i in range( len( feature_selection ) )]

    gu.normalise_features( features, feature_weights )

    mean_features = gu.extract_mean_features( features, sample_array, sample_mask )
    mean_feature_names = feature_names.copy()

    std_of_features = gu.extract_std_of_features( features, sample_array, sample_mask )

    distance_matrix = gu.distance_matrix_from_features( features )

    mean_distance_matrix = gu.distance_matrix_from_features( mean_features )

    mean_features_plus_sample_mask = np.hstack( (np.array( sample_mask )[:, np.newaxis], mean_features) )

    df = gu.array_with_column_titles_to_df( mean_features_plus_sample_mask, ["sample"] + mean_feature_names )

    df.to_csv( ip.output_directory + "TGA/Features/Mean_Features.csv" )

    std_of_features_plus_sample_mask = np.hstack( (np.array( sample_mask )[:, np.newaxis], std_of_features) )

    df = gu.array_with_column_titles_to_df( std_of_features_plus_sample_mask, ["sample"] + mean_feature_names )

    df.to_csv( ip.output_directory + "TGA/Features/Std_of_Features.csv" )

    df = gu.array_with_column_titles_and_label_titles_to_df( mean_distance_matrix, sample_mask, sample_mask )

    df.to_csv( ip.output_directory + "TGA/Features/Distance_Matrix.csv" )

    if plot_specimen_bars:

        for i in range( len( features[0] ) ):

            gu.plot_barchart_of_feature( features[:, i], [f[2] for f in file_data_mask], colour = True, colour_mask = sample_array, filename = ip.output_directory + "TGA/Feature_Bars/Specimen/" + feature_names[i] + ".pdf", savefig = True )

    if plot_mean_bars:

        for i in range( len( mean_features[0] ) ):

            gu.plot_barchart_of_feature( mean_features[:, i], [resin_data.loc[i]["Label"] for i in sample_mask], errorbars = True, std = std_of_features[:, i], colour = True, colour_mask = sample_mask, filename = ip.output_directory + "TGA/Feature_Bars/Mean/" + mean_feature_names[i] + ".pdf", savefig = True )

    if plot_specimen_features:

        gu.plot_features( ip.output_directory, features, feature_names, [f[2] for f in file_data_mask], specimen = True, subdirectory = "TGA/Features/", title = "Specimen_Features.pdf" )

    if plot_mean_features:

        gu.plot_features( ip.output_directory, mean_features, mean_feature_names, [resin_data.loc[i]["Label"] for i in sample_mask], subdirectory = "TGA/Features/", title = "Means_Features.pdf" )

    if plot_specimen_distance_matrix:

        gu.plot_distance_matrix( ip.output_directory, distance_matrix, [f[2] for f in file_data_mask], specimen = True, file_data = file_data_mask, sample_mask = sample_mask, subdirectory = "TGA/Features/", title = "Specimen_Distance_Matrix.pdf" )

    if plot_mean_distance_matrix:

        gu.plot_distance_matrix( ip.output_directory, mean_distance_matrix, [resin_data.loc[i]["Label"] for i in sample_mask], subdirectory = "TGA/Features/", title = "Means_Distance_Matrix.pdf" )

    if plot_specimen_dendrogram:

        gu.plot_dendrogram( ip.output_directory, distance_matrix, [f[2] for f in file_data_mask], specimen = True, subdirectory = "TGA/Features/", title = "Specimen_Dendrogram.pdf" )

    if plot_mean_dendrogram:

        gu.plot_dendrogram( ip.output_directory, mean_distance_matrix, [resin_data.loc[i]["Label"] for i in sample_mask], subdirectory = "TGA/Features/", title = "Means_Dendrogram.pdf" )

def extract_TGA_features_and_analyse_features_2( directory, output_directory, file_data, data, first_derivative_data, second_derivative_data ):

    feature_1 = []

    lower_bound = 400
    upper_bound = 420

    mask = np.where( (first_derivative_data[1] >= lower_bound) & (first_derivative_data[1] <= upper_bound) )[0]

    for i in range( len( file_data ) ):

        feature_1.append( np.mean( first_derivative_data[3][i][mask] ) )

    features = np.array( feature_1 )[:, np.newaxis]

    feature_names = ["TGA_Temp_{}-{}".format( lower_bound, upper_bound )]

    feature_1 = []

    lower_bound = 360
    upper_bound = 380

    mask = np.where( (first_derivative_data[1] >= lower_bound) & (first_derivative_data[1] <= upper_bound) )[0]

    for i in range( len( file_data ) ):

        feature_1.append( np.mean( first_derivative_data[3][i][mask] ) )

    features = np.hstack( (features, np.array( feature_1 )[:, np.newaxis]) )

    feature_names.append( "TGA_Temp_{}-{}".format( lower_bound, upper_bound ) )

    feature_1 = []

    lower_bound = 250
    upper_bound = 270

    mask = np.where( (first_derivative_data[1] >= lower_bound) & (first_derivative_data[1] <= upper_bound) )[0]

    for i in range( len( file_data ) ):

        feature_1.append( np.mean( first_derivative_data[3][i][mask] ) )

    features = np.hstack( (features, np.array( feature_1 )[:, np.newaxis]) )

    feature_names.append( "TGA_Temp_{}-{}".format( lower_bound, upper_bound ) )

    resin_data = gu.get_list_of_resins_data( directory )

    sample, sample_array, samples_present, samples_present_array = gu.sample_data_from_file_data( file_data )

    sample_mask = [11, 14, 10, 4, 13, 21, 23, 18, 22, 20, 2, 3, 17, 16, 19, 1, 15, 12, 6, 5, 7, 9, 8, 24]

    sample_mask = gu.remove_redundant_samples( sample_mask, samples_present )

    feature_selection = [i for i in range( len( feature_names ) )]

    features = features[:, feature_selection]
    feature_names = list( np.array( feature_names )[feature_selection] )

    specimen_mask = gu.produce_mask( sample_array, sample_mask )

    features = features[specimen_mask, :]
    file_data_mask = np.array( file_data )[specimen_mask] # Orders file data according to specimen mask.
    file_data_mask = [[int( f[0] ), int( f[1] ), f[2], f[3]] for f in file_data_mask] # Converts file data so that f[0], etc. are integers and not np.str.
    sample_array = sample_array[specimen_mask]

    feature_weights = [1 for i in range( len( feature_selection ) )]

    gu.normalise_features( features, feature_weights )

    mean_features = gu.extract_mean_features( features, sample_array, sample_mask )
    mean_feature_names = feature_names.copy()

    std_of_features = gu.extract_std_of_features( features, sample_array, sample_mask )

    mean_features_plus_sample_mask = np.hstack( (np.array( sample_mask )[:, np.newaxis], mean_features) )

    df = gu.array_with_column_titles_to_df( mean_features_plus_sample_mask, ["sample"] + mean_feature_names )

    df.to_csv( output_directory + "TGA/Sandbox/Mean_Features.csv" )

    std_of_features_plus_sample_mask = np.hstack( (np.array( sample_mask )[:, np.newaxis], std_of_features) )

    df = gu.array_with_column_titles_to_df( std_of_features_plus_sample_mask, ["sample"] + mean_feature_names )

    df.to_csv( output_directory + "TGA/Sandbox/Std_of_Features.csv" )

    for i in range( len( features[0] ) ):

        gu.plot_barchart_of_feature( features[:, i], [f[2] for f in file_data_mask], colour = True, colour_mask = sample_array, filename = output_directory + "TGA/Sandbox/Feature_Bars/Specimen/" + feature_names[i] + ".pdf", savefig = True )

    for i in range( len( mean_features[0] ) ):

        gu.plot_barchart_of_feature( mean_features[:, i], [resin_data.loc[i]["Label"] for i in sample_mask], errorbars = True, std = std_of_features[:, i], colour = True, colour_mask = sample_mask, filename = output_directory + "TGA/Sandbox/Feature_Bars/Mean/" + mean_feature_names[i] + ".pdf", savefig = True )

def sandbox( directory, output_directory, file_data, data, first_derivative_data, second_derivative_data ):

    perform_extract_TGA_features_and_analyse_features_2 = True

    if perform_extract_TGA_features_and_analyse_features_2:

        extract_TGA_features_and_analyse_features_2( directory, output_directory, file_data, data, first_derivative_data, second_derivative_data )



import pandas as pd
import numpy as np
import re
from scipy.stats import pearsonr
from scipy.stats import spearmanr
import matplotlib.pyplot as plt
import matplotlib.cm as cm
from sklearn.decomposition import PCA
from sklearn.cluster import KMeans
from sklearn.neural_network import MLPClassifier
from sklearn.neural_network import MLPRegressor
from sklearn.model_selection import train_test_split
from sklearn.metrics import mean_squared_error
from math import sqrt
from sklearn.metrics import r2_score
import math
from sklearn import datasets, linear_model

from . import Preprocessing
from .. import Global_Utilities as gu

def compile_full_dataset_of_features( dataset, sample_mask = [] ):

    samples_present = []

    for i in dataset:

        s_p = [int( j ) for j in i.iloc[:, 0].tolist()]
        
        samples_present = samples_present + list( set( s_p) - set( samples_present ) )

    samples_present = sorted( samples_present )

    features = np.zeros( len( samples_present ) )[:, np.newaxis]

    feature_names = []

    for i in dataset:

        feature_names.extend( i.columns[1:] )

        s_p = [int( j ) for j in i.iloc[:, 0].tolist()]

        for j in range( 1, len( i.columns ) ):

            feature = []

            for k in samples_present:

                if k in s_p:

                    feature.append( i.iloc[s_p.index( k ), j] )

                else:

                    feature.append( None )

            features = np.hstack( (features, np.array( feature )[:, np.newaxis]) )

    features = features[:, 1:]

    if sample_mask:

        array_index = [samples_present.index( s ) for s in sample_mask]

        features = features[array_index, :]

        samples_present = sample_mask

    return features, feature_names, samples_present

def produce_full_dataset_of_features( dataset, sample_mask ):

    features = np.zeros( len( sample_mask ) )[:, np.newaxis]

    feature_names = []

    for i in dataset:

        feature_names.extend( i.columns[1:] )

        samples_present = i.iloc[:, 0].tolist()

        for j in range( 1, len( i.columns ) ):

            feature = []

            for k in sample_mask:

                feature.append( i.iloc[samples_present.index( k ), j] )

            features = np.hstack( (features, np.array( feature )[:, np.newaxis]) )

    features = features[:, 1:]

    return features, feature_names

def rank_features( features ):

    rank_features = np.zeros( len( features ) )[:, np.newaxis]

    for i in range( len( features[0] ) ):

        array = features[:, i]

        temp = array.argsort()

        ranks = np.empty_like( temp )

        ranks[temp] = np.arange( len( array ) )

        rank_features = np.hstack( (rank_features, ranks[:, np.newaxis]) )

    rank_features = rank_features[:, 1:]

    return rank_features

# Historic suggestions of things to correlate.

# feature_1_labels = ["FTIR_965-980"]
# feature_2_labels = ["TT_SAB"]
# feature_1_labels = ["PA Content"]
# feature_2_labels = ["TGA_Temp_400-420"]
# feature_1_labels = ["PET Content"]
# feature_2_labels = ["TGA_Temp_360-380"]
# feature_1_labels = ["Irgafos Content"]
# feature_2_labels = ["TGA_Temp_250-270"]
# feature_1_labels = ["DSC_M_Onset"]
# feature_2_labels = ["TT_YM"]
# feature_1_labels = ["TT_SAB"]
# feature_2_labels = ["TT_SHM"]
# feature_1_labels = ["Rhe_Loss_0.10"]
# feature_2_labels = ["FTIR_908"]

def scatterplots( ip ):

    features_df, std_of_features_df, _ = Preprocessing.read_files_and_preprocess( ip, normalised = False, minimal_datasets = True, feature_1 = ip.scatterplot_x, feature_2 = ip.scatterplot_y )

    ip.sample_mask = features_df.index.tolist()

    feature_1 = features_df[ip.scatterplot_x].to_numpy()
    feature_2 = features_df[ip.scatterplot_y].to_numpy()
    std = [std_of_features_df[ip.scatterplot_x].to_numpy(), std_of_features_df[ip.scatterplot_y].to_numpy()]

    if not ip.shiny:

        resin_data = gu.get_list_of_resins_data( ip.directory ) # Obtain the spreadsheet of data for the resins.
        features_metadata = gu.get_features_metadata( ip.directory ) # Obtain the spreadsheet of metadata for the features.

        corr, _ = pearsonr( feature_1, feature_2 )

        print( "The Pearson correlation coefficient is:", corr )

        gu.plot_scatterplot_of_two_features( ip.directory, feature_1, feature_2, features_df.index, [resin_data.loc[i]["Label"] for i in features_df.index], errorbars = True, std = std, line_of_best_fit = False, exponential_fit = False, xlabel = features_metadata.loc[ip.scatterplot_x]["Axis_Label"], ylabel = features_metadata.loc[ip.scatterplot_y]["Axis_Label"], savefig = False, filename = ip.output_directory + "Global/Scatterplots/New_Figure.pdf" )

    return feature_1.tolist(), feature_2.tolist(), std[0].tolist(), std[1].tolist()

def correlation_heatmap( df, spearman = False ):

    correlation_heatmap = np.zeros( (len( df.columns ), len( df.columns )) )

    for i in range( len( df.columns ) ):

        for j in range( i, len( df.columns ) ):

            if spearman:

                correlation, _ = spearmanr( df.iloc[:, i], df.iloc[:, j] )

            else:

                correlation, _ = pearsonr( df.iloc[:, i], df.iloc[:, j] )

            correlation = abs( correlation )

            correlation_heatmap[i][j] = correlation_heatmap[j][i] = correlation

    correlation_df = gu.array_with_column_titles_and_label_titles_to_df( correlation_heatmap, df.columns, df.columns )

    return correlation_df

def pop_columns_from_dataframe( df, features ):

    popped_column = df.pop( features[0] )

    popped_dataframe = popped_column.to_frame()

    for i in range( 1, len( features ) ):

        popped_column = df.pop( features[i] )

        popped_dataframe = popped_dataframe.merge( popped_column.to_frame(), left_index = True, right_index = True )

    return popped_dataframe

def pop_features_from_dataframe( df, features ):

    popped_column = df.pop( features[0] )

    popped_dataframe = popped_column.to_frame()

    for i in range( 1, len( features ) ):

        popped_column = df.pop( features[i] )

        popped_dataframe = popped_dataframe.merge( popped_column.to_frame(), left_index = True, right_index = True )

    df = df.T

    popped_column = df.pop( features[0] )

    for i in range( 1, len( features ) ):

        popped_column = df.pop( features[i] )

    df = df.T

    return df, popped_dataframe

def euclidean_distance_to_virgin( features, sample_mask, virgin_samples, weighting = False ):

    dist = [[] for i in range( len( virgin_samples ) )]

    if weighting:

        features[:, 0] = features[:, 0] * 0.41
        features[:, 1] = features[:, 1] * 0.25
        features[:, 2] = features[:, 2] * 0.12

    for i in range( len( features ) ):

        for j in range( len( virgin_samples ) ):

            dist[j].append( np.linalg.norm( features[i] - features[sample_mask.index( virgin_samples[j] )] ) )

    dist = [np.array( i ) for i in dist]

    return dist

def distance_to_virgin_rank( resin_data, distance_to_virgin, virgin_samples, sample_mask ):

    sample_mask_array = np.array( sample_mask )

    ranks = []

    for i in range( len( virgin_samples ) ):

        temp = distance_to_virgin[i].argsort()

        print( "Virgin " + resin_data.loc[virgin_samples[i]]["Label"] + ":", [resin_data.loc[i]["Label"] for i in sample_mask_array[temp]] )

        rank = np.zeros_like( temp )
        rank[temp] = np.arange( len( sample_mask ) )

        ranks.append( rank )

    sum_ranks = np.zeros_like( temp )

    for i in range( len( virgin_samples ) ):

        sum_ranks += ranks[i]

    temp = sum_ranks.argsort()

    print( "Mean:", [resin_data.loc[i]["Label"] for i in sample_mask_array[temp]] )

def pca( directory, output_directory, shiny, features_df, std_of_features_df ):

    resin_data = gu.get_list_of_resins_data( directory ) # Obtain the spreadsheet of data for the resins.

    pca_of_whole_dataset = False
    pca_of_pca_of_individual_datasets = True

    if pca_of_whole_dataset:

        gu.perform_pca( directory, features_df, [int( i ) for i in features_df.index], std_error = True, std_of_features_df = std_of_features_df, num_components = 2, filename = output_directory + "Global/PCA/Overall.pdf" )

    if pca_of_pca_of_individual_datasets:

        compute_distance_to_virgin = False

        perform_k_means = False

        dataset_names = ["FTIR", "DSC", "TGA", "Rhe", "TT", "Colour", "SHM", "TLS", "ESCR"]

        sample_mask = features_df.index.to_list()

        features_split_by_dataset_df, std_split_by_dataset_df = [], []

        for i, n in enumerate( dataset_names ):

            df = features_df.filter( regex = r"^" + n )

            features_split_by_dataset_df.append( df )

            df = std_of_features_df.filter( regex = r"^" + n )

            std_split_by_dataset_df.append( df )

        pcas, stds = [], []

        for i in range( len( features_split_by_dataset_df ) ):

            if not (features_split_by_dataset_df[i].empty or len( features_split_by_dataset_df[i].columns ) == 1):

                num_components = 2

                pca = PCA( n_components = num_components )
                pca_ft = pca.fit_transform( features_split_by_dataset_df[i] )

                gu.pca_analysis( pca, features_split_by_dataset_df[i] )

                std = [[] for n in range( num_components )]

                for k in range( num_components ):

                    for l in range( len( std_split_by_dataset_df[i].iloc[:, 0] ) ):

                        s = 0

                        for j in range( len( pca.components_[k] ) ):

                            s += abs( pca.components_[k][j] ) * std_split_by_dataset_df[i].iloc[l, j]

                        std[k].append( s )

                std_array = np.array( std ).transpose()

                pcas.append( gu.array_with_column_titles_and_label_titles_to_df( pca_ft, [dataset_names[i] + "_PC1", dataset_names[i] + "_PC2"], sample_mask ) )
                stds.append( gu.array_with_column_titles_and_label_titles_to_df( std_array, [dataset_names[i] + "_PC1", dataset_names[i] + "_PC2"], sample_mask ) )

            elif len( features_split_by_dataset_df[i].columns ) == 1:

                pcas.append( features_split_by_dataset_df[i] )
                stds.append( std_split_by_dataset_df[i] )

            else:

                pcas.append( pd.DataFrame() )
                stds.append( pd.DataFrame() )

        pcas_to_include, stds_to_include = [], []

        for i in range( len( dataset_names ) ):

            if not pcas[i].empty:

                pcas_to_include.append( pcas[i] )
                stds_to_include.append( stds[i] )

        overall_pca = np.zeros( len( sample_mask ) )[:, np.newaxis]
        overall_stds = np.zeros( len( sample_mask ) )[:, np.newaxis]

        feature_names = []

        for i in range( len( pcas_to_include ) ):

            for j in range( len( pcas_to_include[i].columns ) ):

                overall_pca = np.hstack( (overall_pca, pcas_to_include[i].iloc[:, j].to_numpy()[:, np.newaxis]) )
                overall_stds = np.hstack( (overall_stds, stds_to_include[i].iloc[:, j].to_numpy()[:, np.newaxis]) )

                feature_names.append( pcas_to_include[i].columns[j] )

        overall_pca = overall_pca[:, 1:]
        overall_stds = overall_stds[:, 1:]

        num_components = len( overall_pca[0] )

        if len( pcas_to_include ) != 1:

            for i in range( num_components ):

                overall_pca[:, i] =  (overall_pca[:, i] - overall_pca[:, i].min()) / (overall_pca[:, i].max() - overall_pca[:, i].min())
                overall_stds[:, i] =  overall_stds[:, i] / (overall_pca[:, i].max() - overall_pca[:, i].min())

        overall_pca_df = gu.array_with_column_titles_and_label_titles_to_df( overall_pca, feature_names, sample_mask )
        overall_stds_df = gu.array_with_column_titles_and_label_titles_to_df( overall_stds, feature_names, sample_mask )

        if shiny:

            pca_ft_df, pca_std, components, pca_feature_names = gu.perform_pca( directory, overall_pca_df, [int( i ) for i in overall_pca_df.index], std_error = True, std_of_features_df = overall_stds_df, num_components = 2, filename = output_directory + "Global/PCA/Overall.png" )

        else:

            pca_ft_df, pca_std, components, pca_feature_names = gu.perform_pca( directory, overall_pca_df, [int( i ) for i in overall_pca_df.index], std_error = True, std_of_features_df = overall_stds_df, num_components = 2, filename = output_directory + "Global/PCA/Overall.pdf" )

        pca_ft = pca_ft_df.to_numpy()

        if compute_distance_to_virgin:

            virgin_samples = [16, 17, 19]

            virgin_samples = [i for i in virgin_samples if i in sample_mask]

            distance_to_virgin = euclidean_distance_to_virgin( pca_ft, sample_mask, virgin_samples )

            distance_to_virgin_rank( resin_data, distance_to_virgin, virgin_samples, sample_mask )

            # Virgin V6: ['V6', 'V8', 'PCR 6', 'PCR 1', 'PCR 12', 'PCR 5', 'PCR 15', 'PCR 7', 'V7', 'PCR 9', 'PCR 2', 'PCR 8', 'PCR 3', 'PCR 13', 'PCR 14', 'PCR 11', 'PCR 10', 'PCR 21', 'PCR 20', 'PCR 4', 'PCR 18', 'PCR 22', 'PCR 23']
            # Virgin V7: ['V7', 'PCR 15', 'PCR 3', 'V8', 'PCR 1', 'PCR 12', 'V6', 'PCR 5', 'PCR 2', 'PCR 6', 'PCR 14', 'PCR 9', 'PCR 7', 'PCR 4', 'PCR 11', 'PCR 13', 'PCR 10', 'PCR 21', 'PCR 18', 'PCR 20', 'PCR 22', 'PCR 8', 'PCR 23']
            # Virgin V8: ['V8', 'V6', 'V7', 'PCR 15', 'PCR 1', 'PCR 12', 'PCR 3', 'PCR 5', 'PCR 6', 'PCR 2', 'PCR 7', 'PCR 9', 'PCR 14', 'PCR 11', 'PCR 13', 'PCR 8', 'PCR 10', 'PCR 4', 'PCR 21', 'PCR 20', 'PCR 18', 'PCR 22', 'PCR 23']
            # Mean: ['V8', 'V6', 'PCR 15', 'V7', 'PCR 1', 'PCR 12', 'PCR 5', 'PCR 6', 'PCR 3', 'PCR 2', 'PCR 7', 'PCR 9', 'PCR 14', 'PCR 11', 'PCR 13', 'PCR 8', 'PCR 10', 'PCR 4', 'PCR 21', 'PCR 20', 'PCR 18', 'PCR 22', 'PCR 23']

        if perform_k_means:

            kmeans = KMeans( n_clusters = 3, random_state = 0 ).fit( pca_ft[:, [0, 1]] )

            gu.plot_kmeans_plus_pca( pca_ft, kmeans, sample_mask, [resin_data.loc[i]["Label"] for i in sample_mask], savefig = True, xlabel = "First Principal Component", ylabel = "Second Principal Component", filename = output_directory + "Global/Kmeans/Kmeans.png" )

            label_and_cluster_label = zip( [resin_data.loc[i]["Label"] for i in sample_mask], kmeans.labels_ )

            print( "KMeans cluster labels:", list( label_and_cluster_label ) )

        return pca_ft_df.index.tolist(), pca_ft_df[pca_ft_df.columns[0]].tolist(), pca_ft_df[pca_ft_df.columns[1]].tolist(), np.array( pca_std[0] ).tolist(), np.array( pca_std[1] ).tolist(), components[:, 0].tolist(), components[:, 1].tolist(), pca_feature_names

def distance_to_virgin_analysis_based_on_pcas( output_directory, features_df ):

    dataset_names = ["FTIR", "DSC", "TGA", "Rhe", "TT", "Colour"]
    dataset_labels = ["FTIR", "DSC", "TGA", "Rheo", "Mech", "Colour"]

    iter = gu.subset_combinations_for_all_sizes_of_subsets( len( dataset_names ) )

    virgin_samples = [16, 17, 19]

    num_components = 3

    sample_mask = features_df.index.to_list()

    virgin_samples = [i for i in virgin_samples if i in sample_mask]

    features_split_by_dataset_df = []

    for i, n in enumerate( dataset_names ):

        df = features_df.filter( regex = r"^" + n )

        features_split_by_dataset_df.append( df )

    pcas = []

    for i in range( len( features_split_by_dataset_df ) ):

        if not features_split_by_dataset_df[i].empty:

            pca = PCA( n_components = num_components )
            pca_ft = pca.fit_transform( features_split_by_dataset_df[i].to_numpy() )

            pcas.append( gu.array_with_column_titles_and_label_titles_to_df( pca_ft, [dataset_names[i] + "_PC" + str( j ) for j in range( 1, num_components + 1 )], sample_mask ) )

        else:

            pcas.append( pd.DataFrame() )

    iter_dist_to_virgin = []
    iter_names = []

    for it in iter:

        iter_name = ""

        for i in range( len( dataset_names ) ):

            if i in it:

                if iter_name != "":

                    iter_name += "+"

                iter_name += dataset_labels[i]

        iter_names.append( iter_name )

        pcas_to_include = []

        for i in range( len( dataset_names ) ):

            if i in it and not pcas[i].empty:

                pcas_to_include.append( pcas[i] )

        overall_pca = np.zeros( len( sample_mask ) )[:, np.newaxis]

        feature_names = []

        for i in range( len( pcas_to_include ) ):

            for j in range( len( pcas_to_include[i].columns ) ):

                overall_pca = np.hstack( (overall_pca, pcas_to_include[i].iloc[:, j].to_numpy()[:, np.newaxis]) )

                feature_names.append( pcas_to_include[i].columns[j] )

        overall_pca = overall_pca[:, 1:]

        num_components = len( overall_pca[0] )

        for i in range( num_components ):

            overall_pca[:, i] =  (overall_pca[:, i] - overall_pca[:, i].min()) / (overall_pca[:, i].max() - overall_pca[:, i].min())

        overall_pca_df = gu.array_with_column_titles_and_label_titles_to_df( overall_pca, feature_names, sample_mask )

        pca = PCA( n_components = 2 )
        pca_ft = pca.fit_transform( overall_pca_df )

        iter_dist_to_virgin.append( euclidean_distance_to_virgin( pca_ft, sample_mask, virgin_samples ) )

    iter_mean_resin_dist_to_virgin = []

    for i in range( len( iter ) ):

        array = np.array( iter_dist_to_virgin[i] )

        if i == 0 or iter_names[i] == "FTIR+TGA+Mech":

            iter_mean_resin_dist_to_virgin.append( [(array[0][j] + array[1][j] + array[2][j]) / 3 for j in range( len( array[0] ) )] )

    fig, ax = plt.subplots()

    x = np.arange( len( sample_mask ) )
    width = 0.25
    multiplier = 0

    for i in range( 2 ):

        offset = width * multiplier
        rects = ax.bar( x + offset, np.array( iter_mean_resin_dist_to_virgin )[i], width, label = i )
        multiplier += 1

    ax.set_ylabel( "Mean Distance to Virgin" )
    ax.set_xticks( x, sample_mask )
    ax.set_ylim( 0, 2.2 )

    # plt.show()

    plt.close()

    sum_iter_dist_to_virgin = []

    for i in range( len( iter ) ):

        array_1 = np.array( iter_dist_to_virgin[0] )
        array_2 = np.array( iter_dist_to_virgin[i] )

        sum_iter_dist_to_virgin.append( abs( array_1 - array_2 ).sum() / (len( virgin_samples ) * len( sample_mask )) )

    result_dict = {iter_names[i]: sum_iter_dist_to_virgin[i] for i in range( len ( iter_names ) )}

    max_dist = max( result_dict.values() )
    min_dist = min( result_dict.values() )

    dicts, dfs = [], []

    if False:

        # Triplets, horizontal.

        dicts3 = {key: result_dict[key] for key in result_dict if key.count( "+" ) == 2}

        max_dist = max( dicts3.values() )
        min_dist = min( dicts3.values() )

        dicts3 = dict( sorted( dicts3.items(), key = lambda x:x[1] ) )

        df3 = pd.DataFrame.from_dict( dicts3, orient = "index" )

        df3.index = df3.reset_index( drop = True ).index // 10

        df3 = df3.groupby( level = 0 ).apply( lambda x: x[0].reset_index( drop = True ) ).T

        fig, ax = plt.subplots( 2, 1 )

        fig.set_size_inches( 30, 20 )

        im = ax[0].imshow( df3[df3.columns[0]].to_frame().T, vmin = min_dist, vmax = max_dist + 0.07, cmap = cm.plasma )
        im = ax[1].imshow( df3[df3.columns[1]].to_frame().T, vmin = min_dist, vmax = max_dist + 0.07, cmap = cm.plasma )

        for j in range( 10 ):

            pattern = re.compile( r"^(\w+)\+(\w+)\+(\w+)$" )

            ax[0].text( j, 0, pattern.search( list( dicts3.keys() )[j] ).groups()[0] + "\n" + pattern.search( list( dicts3.keys() )[j] ).groups()[1] + "\n" + pattern.search( list( dicts3.keys() )[j] ).groups()[2], ha = "center", va = "center", color = "w", fontsize = 25 )

        for j in range( 10, 20 ):

            pattern = re.compile( r"^(\w+)\+(\w+)\+(\w+)$" )

            ax[1].text( j - 10, 0, pattern.search( list( dicts3.keys() )[j] ).groups()[0] + "\n" + pattern.search( list( dicts3.keys() )[j] ).groups()[1] + "\n" + pattern.search( list( dicts3.keys() )[j] ).groups()[2], ha = "center", va = "center", color = "w", fontsize = 25 )

        ax[0].tick_params( axis = 'x', which = 'both', bottom = False, top = False, labelbottom = False )
        ax[0].tick_params( axis = 'y', which = 'both', left = False, right = False, labelleft = False )
        ax[1].tick_params( axis = 'x', which = 'both', bottom = False, top = False, labelbottom = False )
        ax[1].tick_params( axis = 'y', which = 'both', left = False, right = False, labelleft = False )

        # fig.colorbar( im, orientation = 'vertical' )
        # fig.subplots_adjust( right = 0.6 )

        # plt.tight_layout()

        cbar_ax = fig.add_axes( [0.8, 0.4, 0.04, 0.2] )
        cbar = fig.colorbar( im, cax = cbar_ax, shrink = 0.2 )
        cbar.ax.tick_params( labelsize = 25 )

        plt.subplots_adjust( left = 0.1, bottom = 0.4, right = 0.9, top = 0.6, wspace = 0.4, hspace = 0.4 )

        plt.savefig( output_directory + "Global/Distance_to_Virgin_Analysis/" + "Horizontal.pdf" )

        plt.close()

    if True:

        # Triplets, vertical.

        dicts3 = {key: result_dict[key] for key in result_dict if key.count( "+" ) == 2}

        max_dist = max( dicts3.values() )
        min_dist = min( dicts3.values() )

        dicts3 = dict( sorted( dicts3.items(), key = lambda x:x[1] ) )

        df3 = pd.DataFrame.from_dict( dicts3, orient = "index" )

        df3.index = df3.reset_index( drop = True ).index // 10

        df3 = df3.groupby( level = 0 ).apply( lambda x: x[0].reset_index( drop = True ) ).T

        fig, ax = plt.subplots( 1, 2 )

        fig.set_size_inches( 20, 30 )

        im = ax[0].imshow( df3[df3.columns[0]].to_frame(), vmin = min_dist, vmax = max_dist + 0.07, cmap = cm.plasma )
        im = ax[1].imshow( df3[df3.columns[1]].to_frame(), vmin = min_dist, vmax = max_dist + 0.07, cmap = cm.plasma )

        for j in range( 10 ):

            pattern = re.compile( r"^(\w+)\+(\w+)\+(\w+)$" )

            ax[0].text( 0, j, pattern.search( list( dicts3.keys() )[j] ).groups()[0] + "\n" + pattern.search( list( dicts3.keys() )[j] ).groups()[1] + "\n" + pattern.search( list( dicts3.keys() )[j] ).groups()[2], ha = "center", va = "center", color = "w", fontsize = 25 )

        for j in range( 10, 20 ):

            pattern = re.compile( r"^(\w+)\+(\w+)\+(\w+)$" )

            ax[1].text( 0, j - 10, pattern.search( list( dicts3.keys() )[j] ).groups()[0] + "\n" + pattern.search( list( dicts3.keys() )[j] ).groups()[1] + "\n" + pattern.search( list( dicts3.keys() )[j] ).groups()[2], ha = "center", va = "center", color = "w", fontsize = 25 )

        ax[0].tick_params( axis = 'x', which = 'both', bottom = False, top = False, labelbottom = False )
        ax[0].tick_params( axis = 'y', which = 'both', left = False, right = False, labelleft = False )
        ax[1].tick_params( axis = 'x', which = 'both', bottom = False, top = False, labelbottom = False )
        ax[1].tick_params( axis = 'y', which = 'both', left = False, right = False, labelleft = False )

        # fig.colorbar( im, orientation = 'vertical' )
        # fig.subplots_adjust( right = 0.6 )

        # plt.tight_layout()

        cbar_ax = fig.add_axes( [0.66, 0.22, 0.06, 0.56] )
        cbar = fig.colorbar( im, cax = cbar_ax, shrink = 1 )
        cbar.ax.tick_params( labelsize = 25 )

        plt.subplots_adjust( left = 0.4, bottom = 0.1, right = 0.6, top = 0.9, wspace = 0.4, hspace = 0.4 )

        plt.savefig( output_directory + "Global/Distance_to_Virgin_Analysis/" + "Vertical.pdf" )

        plt.close()

    if False:

        searches = ["FTIR", "DSC", "TGA", "Rheo", "Mech", "Colour"]

        for i, s in enumerate( searches ):

            pattern = re.compile( s )

            dicts.append( {key: result_dict[key] for key in result_dict if pattern.search( key ) != None} )

            dfs.append( pd.DataFrame.from_dict( dicts[i], orient = "index" ) )

            dfs[i].rename( columns = {dfs[i].columns[0]:s}, inplace = True )

        fig, ax = plt.subplots( 1, 6 )

        fig.set_size_inches( 20, 20 )

        for i in range( len( searches ) ):

            im = ax[i].imshow( dfs[i], vmin = 0, vmax = max_dist, cmap = cm.plasma )

            ax[i].set_xticks( np.arange( 0, len( dfs[i].columns ), 1 ) )
            ax[i].set_yticks( np.arange( 0, len( dfs[i].index ), 1 ) )

            ax[i].set_xticklabels( dfs[i].columns, rotation = 270, fontsize = 20 )
            ax[i].set_yticklabels( dfs[i].index, fontsize = 10 )

            # for j in range( len( dfs[i].index ) ):
            #
            #     pattern = re.compile( r"^(\w+)\+(\w+)\+(\w+)$" )
            #
            #     ax[i].text( 0, j, pattern.search( dfs[i].index[j] ).groups()[0] + "\n" + pattern.search( dfs[i].index[j] ).groups()[1] + "\n" + pattern.search( dfs[i].index[j] ).groups()[2], ha = "center", va = "center", color = "w" )

            ax[i].invert_yaxis()

        plt.tight_layout()

        plt.savefig( output_directory + "Global/Distance_to_Virgin_Analysis/" + "Plot.pdf" )

        plt.close()

def manual_ml( directory, ip, features_df ):

    resin_data = gu.get_list_of_resins_data( directory ) # Obtain the spreadsheet of data for the resins.

    ml_features = pd.DataFrame( features_df["FTIR_777"] )
    ml_features = ml_features.merge( features_df["DSC_HFM_160"], left_index = True, right_index = True )
    ml_features = ml_features.merge( features_df["TGA_Temp_380-400"], left_index = True, right_index = True )
    ml_features = ml_features.merge( features_df["Rhe_Crossover"], left_index = True, right_index = True )

    scalars = [-1, -0.7, -0.3, 0, 0.3, 0.7, 1]

    a, b, c, d = 0, 0, 0, 0

    max_pearson = 0

    for i in scalars:

        for j in scalars:

            for k in scalars:

                for l in scalars:

                    if i == 0 and j == 0 and k == 0 and l == 0:

                        continue

                    ml_copy = ml_features.copy()

                    ml_copy[ml_copy.columns[0]] = ml_copy[ml_copy.columns[0]].apply( lambda x: x * i )
                    ml_copy[ml_copy.columns[1]] = ml_copy[ml_copy.columns[1]].apply( lambda x: x * j )
                    ml_copy[ml_copy.columns[2]] = ml_copy[ml_copy.columns[2]].apply( lambda x: x * k )
                    ml_copy[ml_copy.columns[3]] = ml_copy[ml_copy.columns[3]].apply( lambda x: x * l )

                    sum_ml = ml_copy.sum( axis = 1 )

                    pearson, _ = pearsonr( sum_ml, features_df["TT_SAB"].tolist() )

                    if abs( pearson ) > max_pearson:

                        max_pearson = abs( pearson )
                        a, b, c, d = i, j, k, l

    print( a, b, c, d, max_pearson )

    ml_copy = ml_features.copy()

    ml_copy[ml_copy.columns[0]] = ml_copy[ml_copy.columns[0]].apply( lambda x: x * a )
    ml_copy[ml_copy.columns[1]] = ml_copy[ml_copy.columns[1]].apply( lambda x: x * b )
    ml_copy[ml_copy.columns[2]] = ml_copy[ml_copy.columns[2]].apply( lambda x: x * c )
    ml_copy[ml_copy.columns[3]] = ml_copy[ml_copy.columns[3]].apply( lambda x: x * d )

    sum_ml = ml_copy.sum( axis = 1 )

    feature_1 = sum_ml.to_numpy()
    feature_2 = features_df["TT_SAB"].to_numpy()

    # gu.plot_scatterplot_of_two_features( feature_1, feature_2, ip.sample_mask, [resin_data.loc[i]["Label"] for i in features_df.index] )

    X = ml_features.to_numpy()
    y = features_df["TT_SAB"].to_numpy()

    training_resins = [1, 2, 3, 4, 5, 6, 8, 10, 12, 14, 16, 18, 20, 22, 23]
    test_resins = [7, 9, 11, 13, 15, 17, 19, 21]

    split_train = [ml_features.index.to_list().index( i ) for i in training_resins]
    split_test = [ml_features.index.to_list().index( i ) for i in test_resins]

    X_train = X[split_train]
    X_test = X[split_test]
    y_train = y[split_train]
    y_test = y[split_test]

    regr = linear_model.LinearRegression()

    regr.fit( X_train, y_train )

    y_pred = regr.predict( X_test )

    print( "Coefficients: ", regr.coef_ ) # The coefficients.
    print( "Mean squared error: %.2f" % mean_squared_error( y_test, y_pred ) ) # The mean squared error.
    print( "Coefficient of determination: %.2f" % r2_score( y_test, y_pred ) ) # The coefficient of determination: 1 is perfect prediction.

    gu.plot_scatterplot_of_two_features( y_test, y_pred, test_resins, [resin_data.loc[i]["Label"] for i in test_resins] )

def pca_ml( directory, ip, features_df ):

    resin_data = gu.get_list_of_resins_data( directory ) # Obtain the spreadsheet of data for the resins.

    num_components = 5

    X, _, _, _ = gu.perform_pca( directory, features_df, ip.sample_mask, num_components = num_components, analysis_of_pca = False )

    # gu.plot_df_heatmap( X )

    X, y = X.to_numpy(), features_df["TT_SAB"].to_numpy()

    number_repeats = 1

    coefficients = []
    mse = []
    cod = []

    for r in range( number_repeats ):

        training_resins = [1, 2, 3, 4, 5, 6, 8, 10, 12, 14, 16, 18, 20, 22, 23]
        test_resins = [7, 9, 11, 13, 15, 17, 19, 21]

        split_train = [features_df.index.to_list().index( i ) for i in training_resins]
        split_test = [features_df.index.to_list().index( i ) for i in test_resins]

        # X_train, X_test, y_train, y_test = train_test_split( X, y, test_size = 7 )

        X_train = X[split_train]
        X_test = X[split_test]
        y_train = y[split_train]
        y_test = y[split_test]

        regr = linear_model.LinearRegression()

        regr.fit( X_train, y_train )

        y_pred = regr.predict( X_test )

        coefficients.append( regr.coef_ )
        mse.append( mean_squared_error( y_test, y_pred ) )
        cod.append( r2_score( y_test, y_pred ) )

        print( "Coefficients: ", regr.coef_ ) # The coefficients.
        print( "Mean squared error: %.2f" % mean_squared_error( y_test, y_pred ) ) # The mean squared error.
        print( "Coefficient of determination: %.2f" % r2_score( y_test, y_pred ) ) # The coefficient of determination: 1 is perfect prediction.

        gu.plot_scatterplot_of_two_features( y_test, y_pred, test_resins, [resin_data.loc[i]["Label"] for i in test_resins] )

    mean_mse = np.array( mse ).mean()
    mean_cod = np.array( cod ).mean()
    c_coefficients = []

    for j in range( num_components ):

        c_coefficients.append( [coefficients[i][j] for i in range( number_repeats )] )

    mean_coefficients = [np.array( c_coefficients[i] ).mean() for i in range( num_components )]

    print( "Mean MSE: ", mean_mse )
    print( "Mean cod: ", mean_cod )
    print( "Mean coefficients: ", mean_coefficients )

    # Output: Mean MSE:  0.07447263900017585
    # Mean cod:  0.3498298501594448
    # Mean coefficients:  [-0.15599656149109054, -0.25995542978716757, 0.10756368212119027, -0.08592974744840264, 0.21583468703827902]

def common_value( list_1, list_2, i ):

    if type( list_1[i] ) == str or type( list_2[i] ) == str:

        return False

    elif math.isnan( list_1[i] ) or math.isnan( list_2[i] ):

        return False

    else:

        return True

def plot_two_columns_of_df( output_directory, df, c1, c2 ):

    list_1 = df[df.columns[c1]].tolist()
    list_2 = df[df.columns[c2]].tolist()

    list_3 = [list_1[x] for x in range( len( list_1 ) ) if common_value( list_1, list_2, x )]
    list_4 = [list_2[x] for x in range( len( list_2 ) ) if common_value( list_2, list_1, x )]

    fig, ax = plt.subplots()

    ax.plot( list_4, list_3, 'rx' )

    ax.set_xscale( 'log' )
    ax.set_yscale( 'log' )
    ax.set_title( "Zero-shear Viscosity")
    ax.set_xlabel( "Zero-Shear Viscosity " )
    ax.set_ylabel( "MFI (190°, 2.16kg)" )
    plt.tight_layout()

    array_3 = np.array( list_3 )
    array_4 = np.array( list_4 )

    m, b = np.polyfit( np.log( array_4 ), np.log( array_3 ), 1 )

    array_4 = np.sort( array_4 )

    plt.plot( array_4, array_4 ** m * np.exp( b ) )

    for i in range( len( df.index.values ) ):

        if df.iat[i, 4] == 1:

            ax.errorbar( df.iat[i, 8], df.iat[i, 3], 0.5 * (df.iat[i, 6] - df.iat[i, 5]), capsize = 5 )

    output_file = df.columns[c1][:3] + df.columns[c2][:3] + ".png"

    plt.savefig( output_directory + output_file )

def mfi_vs_zero_shear( directory ):

    df = pd.read_csv( directory + "rhdpe_dataset.csv" )

    df.drop( df.columns[0], axis = 1, inplace = True )

    # df.drop( [18, 21, 23], inplace = True )

    plot_two_columns_of_df( directory + "Global/Output/Sandbox/MFI/", df, 3, 8 )

def mlpregressor( features_df ):

    target_column = ['TT_SAB']
    predictors = list( set( list( features_df.columns ) ) - set( target_column ) )

    predictors = ["FTIR_997", "Rhe_Crossover", "TT_SHM"]
    features_df[predictors] = features_df[predictors]/features_df[predictors].max()

    X = features_df[predictors].values
    y = features_df[target_column].values

    X_train, X_test, y_train, y_test = train_test_split( X, y, test_size = 0.20, random_state = 40 )

    mlp = MLPRegressor( hidden_layer_sizes=( 8, 8, 8 ), activation = 'relu', solver = 'adam', max_iter = 500 )
    mlp.fit( X_train, y_train.ravel() )

    predict_train = mlp.predict( X_train )
    predict_test = mlp.predict( X_test )

    pred = mlp.predict(X_test)

    # Calculate accuracy and error metrics

    test_set_rsquared = mlp.score( X_test, y_test )
    test_set_rmse = np.sqrt( mean_squared_error( y_test, pred ) )

    # Print R_squared and RMSE value

    print( 'R_squared value: ', test_set_rsquared )
    print( 'RMSE: ', test_set_rmse )

    plt.plot( predict_train, y_train, 'o' )
    plt.plot( pred, y_test, 'o' )
    plt.show()

from sklearn.cross_decomposition import PLSRegression

def pls( features_df ):
    '''https://towardsdatascience.com/partial-least-squares-f4e6714452a'''

    X_colnames = features_df.columns[:-7].to_list() + features_df.columns[-3:].to_list()
    Y_colnames = features_df.columns[-7:-3].to_list()

    X = features_df[X_colnames].values
    Y = features_df[Y_colnames].values

    sab_plot = []
    uts_plot = []
    ym_plot = []
    shm_plot = []

    for n_comp in range( 1, 21 ):

      my_plsr = PLSRegression( n_components = n_comp, scale = True )
      my_plsr.fit( X, Y )
      preds = my_plsr.predict( X )

      sab_rmse = sqrt( mean_squared_error( Y[:,0] , preds[:,0] ) )
      uts_rmse = sqrt( mean_squared_error( Y[:,1] , preds[:,1] ) )
      ym_rmse = sqrt( mean_squared_error( Y[:,2] , preds[:,2] ) )
      shm_rmse = sqrt( mean_squared_error( Y[:,3] , preds[:,3] ) )

      sab_plot.append( sab_rmse )
      uts_plot.append( uts_rmse )
      ym_plot.append( ym_rmse )
      shm_plot.append( shm_rmse )

    # Create the three plots using matplotlib
    fig, axs = plt.subplots(1,4)

    axs[0].plot( range( 1, 21 ), sab_plot )
    axs[1].plot( range( 1, 21 ), uts_plot )
    axs[2].plot( range( 1, 21 ), ym_plot )
    axs[3].plot( range( 1, 21 ), shm_plot )

    # plt.show()


    best_model = PLSRegression( n_components = 10, scale = True )
    best_model.fit( X, Y )
    test_preds = best_model.predict( X )
    print( r2_score( Y, test_preds ) )

import itertools
from sklearn.linear_model import LinearRegression

def escr_predictor( directory, features_df, std_of_features_df ):

    # [1.12733926] -0.1283258558485133

    resin_data = gu.get_list_of_resins_data( directory )

    # target_escr = features_df["ESCR_50%Failure"]
    #
    # X_escr = pd.DataFrame( features_df["SHM"] )
    #
    # reg = LinearRegression().fit( X_escr, target_escr )
    #
    # score = reg.score( X_escr, target_escr )
    #
    # print( score )
    # print( reg.coef_, reg.intercept_)
    #
    # prediction = reg.predict( X_escr )
    #
    # gu.plot_scatterplot_of_two_features( directory, prediction, target_escr.to_numpy(), features_df.index, [resin_data.loc[i]["Label"] for i in features_df.index], errorbars = False, std = "", line_of_best_fit = False, exponential_fit = False, xlabel = "Prediction", ylabel = "Target", savefig = False, filename = "" )
    #
    # plt.show()
    #
    # plt.close()

    target = features_df["SHM"]

    X = features_df.drop( ["SHM", "TT_SHM"], axis = 1 )
    # X = features_df.drop( ["SHM"], axis = 1 )

    print( X.shape )

    combinations = itertools.combinations( [i for i in range( 23 )], 3 )

    i = 0
    score = 0
    best_c = None

    # for c in combinations:
    #
    #     i += 1
    #
    #     if (i > 100):
    #
    #         pass
    #         # break
    #
    #     X1 = X.iloc[:, list( c )]
    #
    #     reg = LinearRegression().fit( X1, target )
    #
    #     score_1 = reg.score( X1, target )
    #
    #     if score_1 > score:
    #
    #         score = score_1
    #         best_c = c

    best_c = (2, 11, 14)

    print( score, best_c, X.iloc[:, list( best_c )].columns )

    X1 = X.iloc[:, list( best_c )]

    reg = LinearRegression().fit( X1, target )

    prediction = reg.predict( X1 )

    prediction_escr = prediction * 1.12 - 0.13

    unnormalised_shm = prediction * 9 + 5
    unnormalised_escr = prediction_escr * 80 + 16

    array = np.hstack( (unnormalised_shm[:, np.newaxis], unnormalised_escr[:, np.newaxis]) )

    np.savetxt( "/Users/philsmith/Documents/Postdoc/rHDPE_Data_Analysis/Output/Global/Sandbox/ESCR_Prediction/ESCR_Prediction.csv", array, delimiter = ",", fmt = "%.5f" )

    gu.plot_scatterplot_of_two_features( directory, prediction, target.to_numpy(), features_df.index, [resin_data.loc[i]["Label"] for i in features_df.index], errorbars = False, std = "", line_of_best_fit = False, exponential_fit = False, xlabel = "Prediction", ylabel = "Target", savefig = False, filename = "" )

def feature_regression( directory, features_df ):

    # PP and CaCO3 vs SAB.

    # feature_1_labels = ["FTIR_868-885", "FTIR_965-980"]
    # feature_2_labels = ["TT_SAB"]

    resin_data = gu.get_list_of_resins_data( directory )

    target_feature = "ESCR_50%Failure"
    regression_features = ["SHM", "TT_SAB"]

    target = features_df[target_feature]
    features_df = features_df.drop( target_feature, axis = 1 )

    X = features_df[regression_features]

    reg = LinearRegression().fit( X, target )

    prediction = reg.predict( X )

    for i in range( len( regression_features ) ):

        print( regression_features[i], "has coefficient", reg.coef_[i] )

    print( "Intercept:", reg.intercept_ )

    gu.plot_scatterplot_of_two_features( directory, prediction, target.to_numpy(), features_df.index, [resin_data.loc[i]["Label"] for i in features_df.index], errorbars = False, std = "", line_of_best_fit = False, exponential_fit = False, xlabel = "Prediction", ylabel = "Target", savefig = False, filename = "" )

def sandbox( directory, features_df, std_of_features_df ):

    perform_mfi_vs_zero_shear = False
    perform_mlpregressor = False
    perform_pls = False
    perform_escr_predictor = False
    perform_feature_regression = True

    if perform_mfi_vs_zero_shear:

        mfi_vs_zero_shear( directory )

    if perform_mlpregressor:

        mlpregressor( features_df )

    if perform_pls:

        pls( features_df )

    if perform_escr_predictor:

        escr_predictor( directory, features_df, std_of_features_df )

    if perform_feature_regression:

        feature_regression( directory, features_df )